import base64
import secrets

from lilya.exceptions import PermissionDenied
from lilya.protocols.permissions import PermissionProtocol
from lilya.requests import Request
from lilya.types import ASGIApp, Receive, Scope, Send


class BasicAuthAccess(PermissionProtocol):
    def __init__(
        self, app: ASGIApp, *, username: str = "admin", password: str, print_pw: bool = False
    ) -> None:
        self.app = app
        self.basic_string = base64.b64encode(f"{username}:{password}".encode()).decode()
        if print_pw:
            print("The admin panel password is:", password)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope=scope, receive=receive, send=send)
        if "Authorization" not in request.headers:
            raise PermissionDenied(
                status_code=401,
                headers={"WWW-Authenticate": 'Basic realm="Admin panel", charset="UTF-8"'},
            )

        auth = request.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != "basic":
                raise PermissionDenied(
                    status_code=401,
                    headers={"WWW-Authenticate": 'Basic realm="Admin panel", charset="UTF-8"'},
                )
        except ValueError as exc:
            raise PermissionDenied(
                status_code=401,
                headers={"WWW-Authenticate": 'Basic realm="Admin panel", charset="UTF-8"'},
            ) from exc

        if not secrets.compare_digest(credentials, self.basic_string):
            raise PermissionDenied(
                status_code=401,
                headers={"WWW-Authenticate": 'Basic realm="Admin panel", charset="UTF-8"'},
            )

        await self.app(scope, receive, send)



from typing import cast

from lilya.context import session


def add_message(level: str, message: str) -> None:
    """
    Stores a message in the session for rendering in the next request.
    Level can be: success, info, warning, error.
    """
    if not hasattr(session, "messages"):
        session.messages = []
    cast(list, session.messages).append({"level": level, "text": message})


def get_messages(peek: bool = False) -> list:
    """
    Retrieves and clears messages from the session.
    """
    if not hasattr(session, "messages"):
        return []
    messages = cast(list, session.messages)
    if peek:
        return messages
    _messages = messages.copy()
    messages.clear()
    return _messages



from __future__ import annotations

import secrets
from typing import Any

import click

import edgy
from edgy.conf import settings


@click.option(
    "-p",
    "--port",
    type=int,
    default=8000,
    help="Port to run the development server.",
    show_default=True,
)
@click.option(
    "--host",
    type=str,
    default="localhost",
    help="Server host. Typically localhost.",
    show_default=True,
)
@click.option(
    "--admin-path",
    type=str,
    default=None,
    help='The sub-path under which the admin is available. Defaults to "/admin" if unset in settings',
    show_default=True,
)
@click.option(
    "--auth-name",
    type=str,
    default="admin",
    help="Basic auth name.",
    show_default=True,
)
@click.option(
    "--auth-pw",
    type=str,
    default=None,
    help="Basic auth password. By default autogenerated.",
    show_default=False,
)
@click.option(
    "--debug",
    help="Start the application in debug mode.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--create-all",
    help="Create all models when not existent.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--log-level",
    type=str,
    default="info",
    help="What log level should uvicorn run.",
    show_default=True,
)
@click.command(name="runserver")
def admin_serve(
    port: int,
    host: str,
    debug: bool,
    create_all: bool,
    log_level: str,
    auth_name: str,
    auth_pw: str | None,
    admin_path: str | None,
) -> None:
    """Starts the Edgy admin development server.

    The --app can be passed in the form of <module>.<submodule>:<app> or be set
    as environment variable LILYA_DEFAULT_APP.

    Alternatively, if none is passed, Lilya will perform the application discovery.

    It is strongly advised not to run this command in any pther environment but developmentyping.
    This was designed to facilitate the development environment and should not be used in pr

    How to run: `edgy admin_serve`
    """
    try:
        from lilya.apps import Lilya
        from lilya.middleware import DefineMiddleware
        from lilya.middleware.sessions import SessionMiddleware
        from lilya.permissions import DefinePermission
        from lilya.routing import Include
    except ImportError:
        raise RuntimeError("Lilya needs to be installed to run admin_serve.") from None
    try:
        import jinja2  # noqa
    except ImportError:
        raise RuntimeError("Jinja2 needs to be installed to run admin_serve.") from None
    try:
        import uvicorn
    except ImportError:
        raise RuntimeError("Uvicorn needs to be installed to run admin_serve.") from None
    from edgy.contrib.admin.permissions import BasicAuthAccess

    old_instance = edgy.monkay.instance
    if admin_path is not None:
        edgy.monkay.settings.admin_config.admin_prefix_url = admin_path
    if edgy.monkay.settings.admin_config.admin_prefix_url is None:
        edgy.monkay.settings.admin_config.admin_prefix_url = "/admin"

    if old_instance is None:
        raise RuntimeError(
            'You need to specify an app which registry is used. For experimenting use: "tests.cli.main"'
        )
    from edgy.contrib.admin.application import create_admin_app

    print_pw = False
    if auth_pw is None:
        auth_pw = secrets.token_urlsafe()
        print_pw = True
    admin_app = create_admin_app()
    routes = [
        Include(
            path=settings.admin_config.admin_prefix_url,
            app=admin_app,
            permissions=[
                DefinePermission(
                    BasicAuthAccess, print_pw=print_pw, username=auth_name, password=auth_pw
                )
            ],
            middleware=[
                DefineMiddleware(
                    SessionMiddleware,
                    secret_key=settings.admin_config.SECRET_KEY,
                    session_cookie="admin_session",
                ),
            ],
        ),
    ]
    if old_instance.app is not None:
        routes.append(Include(path="/", app=old_instance.app))
    app: Any = Lilya(
        routes=routes,
    )
    if debug:
        app.debug = True
        admin_app.debug = True
    app = old_instance.registry.asgi(app)
    if create_all:
        edgy.run_sync(old_instance.registry.create_all())
    edgy.monkay.set_instance(edgy.Instance(registry=old_instance.registry, app=app))

    uvicorn.run(
        app=app,
        port=port,
        host=host,
        reload=False,
        lifespan="on",
        log_level=log_level,
    )



from __future__ import annotations

from base64 import urlsafe_b64decode, urlsafe_b64encode
from contextlib import suppress
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import orjson
from lilya.exceptions import ImproperlyConfigured
from lilya.requests import Request
from lilya.templating import Jinja2Template

from edgy.conf import settings
from edgy.contrib.admin.utils.messages import get_messages
from edgy.core.db.fields.file_field import ConcreteFileField
from edgy.exceptions import ObjectNotFound


@lru_cache(maxsize=1)
def get_templates() -> Jinja2Template:
    templates = Jinja2Template(
        directory=[
            *settings.admin_config.admin_extra_templates,
            Path(__file__).resolve().parent / "templates",
        ]
    )
    templates.env.globals["getattr"] = getattr
    templates.env.globals["isinstance"] = isinstance
    return templates


if TYPE_CHECKING:
    from edgy.core.db.models.model import Model


class AdminMixin:
    @property
    def templates(self) -> Jinja2Template:
        return get_templates()

    def get_admin_prefix_url(self, request: Request) -> str:
        if settings.admin_config.admin_prefix_url is not None:
            return settings.admin_config.admin_prefix_url
        return str(request.path_for("admin")).rstrip("/")

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:  # noqa
        context = {}
        user: Any = None
        with suppress(ImproperlyConfigured):
            user = request.user
        context.update(
            {
                "user": user,
                "isinstance": isinstance,
                "ConcreteFileField": ConcreteFileField,
                "title": settings.admin_config.title,
                "dasboard_title": settings.admin_config.dashboard_title,
                "menu_title": settings.admin_config.menu_title,
                "favicon": settings.admin_config.favicon,
                "sidebar_bg_colour": settings.admin_config.sidebar_bg_colour,
                "url_prefix": self.get_admin_prefix_url(request),
                "messages": get_messages(),
                "create_object_pk": self.create_object_pk,
            }
        )
        return context

    def create_object_pk(self, instance: Model) -> str:
        """
        Extracts the object ID from the request's path parameters.

        Assumes the object ID is present in the path parameters under the key "id".

        Args:
            request: The incoming Starlette Request object.

        Returns:
            The object ID as a string.
        """
        return urlsafe_b64encode(
            orjson.dumps({name: getattr(instance, name) for name in instance.pknames})
        ).decode()

    def get_object_pk(self, request: Request) -> dict:
        """
        Extracts the object ID from the request's path parameters.

        Assumes the object ID is present in the path parameters under the key "id".

        Args:
            request: The incoming Starlette Request object.

        Returns:
            The object ID as dict.
        """
        try:
            result = cast(dict, orjson.loads(urlsafe_b64decode(request.path_params.get("id"))))
            if not isinstance(result, dict):
                raise ObjectNotFound()
            return result
        except orjson.JSONDecodeError:
            raise ObjectNotFound() from None



import copy
import datetime
import decimal
import enum
import ipaddress
import uuid
import warnings
from collections.abc import Callable
from enum import Enum, EnumMeta
from re import Pattern
from secrets import compare_digest
from typing import TYPE_CHECKING, Annotated, Any, Optional, cast

import orjson
import pydantic
import sqlalchemy
from monkay import Monkay
from pydantic.networks import AnyUrl, EmailStr, IPvAnyAddress
from sqlalchemy.dialects import oracle

from edgy.core.db.context_vars import CURRENT_PHASE
from edgy.core.db.fields._internal import IPAddress
from edgy.core.db.fields.base import Field
from edgy.core.db.fields.factories import FieldFactory
from edgy.core.db.fields.types import BaseFieldType
from edgy.exceptions import FieldDefinitionError

from .mixins import AutoNowMixin as _AutoNowMixin
from .mixins import IncrementOnSaveBaseField, TimezonedField
from .place_holder_field import PlaceholderField as _PlaceholderField

if TYPE_CHECKING:
    import zoneinfo


CLASS_DEFAULTS = ["cls", "__class__", "kwargs"]


class CharField(FieldFactory, str):
    """String field representation that constructs the Field class and populates the values"""

    field_type = str

    def __new__(
        cls,
        *,
        min_length: int | None = None,
        regex: str | Pattern = None,
        pattern: str | Pattern = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        if pattern is None:
            pattern = regex
        del regex
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        max_length = kwargs.get("max_length", 0)
        if max_length is not None and max_length <= 0:
            raise FieldDefinitionError(detail=f"'max_length' is required for {cls.__name__}")

        min_length = kwargs.get("min_length")
        pattern = kwargs.get("regex")

        assert min_length is None or isinstance(min_length, int)
        assert max_length is None or isinstance(max_length, int)
        assert pattern is None or isinstance(pattern, str | Pattern)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        max_length: int | None = kwargs.get("max_length")
        return (
            sqlalchemy.Text(collation=kwargs.get("collation"))
            if max_length is None
            else sqlalchemy.String(length=max_length, collation=kwargs.get("collation"))
        )


class TextField(CharField):
    """String representation of a text field which means no max_length required"""

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", None)
        super().validate(kwargs)


class IntegerField(FieldFactory, int):
    """
    Integer field factory that construct Field classes and populated their values.
    """

    field_type = int
    field_bases = (IncrementOnSaveBaseField,)

    def __new__(
        cls,
        *,
        ge: int | float | decimal.Decimal | None = None,
        gt: int | float | decimal.Decimal | None = None,
        le: int | float | decimal.Decimal | None = None,
        lt: int | float | decimal.Decimal | None = None,
        multiple_of: int | None = None,
        increment_on_save: int = 0,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Integer()

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        increment_on_save = kwargs.get("increment_on_save", 0)
        if (
            increment_on_save == 0
            and kwargs.get("primary_key", False)
            and "autoincrement" not in kwargs
        ):
            warnings.warn(
                (
                    "Not setting autoincrement on an Integer field with `primary_key=True` is deprecated."
                    "We change the default to False in future."
                ),
                DeprecationWarning,
                stacklevel=4,
            )
            kwargs["autoincrement"] = True
        if increment_on_save != 0:
            if kwargs.get("autoincrement"):
                raise FieldDefinitionError(
                    detail="'autoincrement' is incompatible with 'increment_on_save'"
                )
            if kwargs.get("null"):
                raise FieldDefinitionError(
                    detail="'null' is incompatible with 'increment_on_save'"
                )
            kwargs.setdefault("read_only", True)
            kwargs["inject_default_on_partial_update"] = True


class FloatField(FieldFactory, float):
    """Representation of a int32 and int64"""

    field_type = float

    def __new__(
        cls,
        *,
        max_digits: int | None = None,
        ge: int | float | decimal.Decimal | None = None,
        gt: int | float | decimal.Decimal | None = None,
        le: int | float | decimal.Decimal | None = None,
        lt: int | float | decimal.Decimal | None = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        # pydantic doesn't support max_digits for float, so rename it
        if max_digits is not None:
            kwargs.setdefault("precision", max_digits)
        del max_digits
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        precision: int | None = kwargs.get("precision")
        if precision is None:
            return sqlalchemy.Float(asdecimal=False)
        return sqlalchemy.Float(precision=precision, asdecimal=False).with_variant(
            oracle.FLOAT(binary_precision=round(precision / 0.30103), asdecimal=False),  # type: ignore
            "oracle",
        )


class BigIntegerField(IntegerField):
    """Representation of big integer field"""

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        if kwargs.get("autoincrement", False):
            kwargs.setdefault("skip_reflection_type_check", True)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        # sqlite special we cannot have a big IntegerField as PK
        if kwargs.get("autoincrement"):
            return sqlalchemy.BigInteger().with_variant(sqlalchemy.Integer, "sqlite")
        return sqlalchemy.BigInteger()


class SmallIntegerField(IntegerField):
    """Represents a small integer field"""

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.SmallInteger()


class DecimalField(FieldFactory, decimal.Decimal):
    field_type = decimal.Decimal

    def __new__(
        cls,
        *,
        ge: int | float | decimal.Decimal | None = None,
        gt: int | float | decimal.Decimal | None = None,
        le: int | float | decimal.Decimal | None = None,
        lt: int | float | decimal.Decimal | None = None,
        multiple_of: int | decimal.Decimal | None = None,
        max_digits: int | None = None,
        decimal_places: int | None = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Numeric(
            precision=kwargs.get("max_digits"), scale=kwargs.get("decimal_places"), asdecimal=True
        )

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)

        decimal_places = kwargs.get("decimal_places")
        if decimal_places is None or decimal_places < 0:
            raise FieldDefinitionError("decimal_places are required for DecimalField")


# in python it is not possible to subclass bool. So only use bool for type checking.
class BooleanField(FieldFactory, cast(bool, int)):
    """Representation of a boolean"""

    field_type = bool

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Boolean()


class DateTimeField(_AutoNowMixin, datetime.datetime):
    """Representation of a datetime field"""

    field_type = datetime.datetime
    field_bases = (TimezonedField, Field)

    def __new__(
        cls,
        *,
        auto_now: bool | None = False,
        auto_now_add: bool | None = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        force_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        remove_timezone: bool = False,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        kwargs.setdefault("with_timezone", not remove_timezone)
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        with_timezone: bool = kwargs.get("with_timezone", True)
        return sqlalchemy.DateTime(with_timezone)

    @classmethod
    def get_default_values(
        cls,
        field_obj: Field,
        field_name: str,
        cleaned_data: dict[str, Any],
        original_fn: Any = None,
    ) -> Any:
        phase = CURRENT_PHASE.get()
        if field_obj.auto_now_add and phase == "prepare_update":
            return {}
        return original_fn(field_name, cleaned_data)


class DateField(_AutoNowMixin, datetime.date):
    """Representation of a date field"""

    field_type = datetime.date
    field_bases = (TimezonedField, Field)

    def __new__(
        cls,
        *,
        auto_now: bool | None = False,
        auto_now_add: bool | None = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        force_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        # the datetimes lose the information anyway
        kwargs["remove_timezone"] = False

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Date()


class DurationField(FieldFactory, datetime.timedelta):
    """Representation of a time field"""

    field_type = datetime.timedelta

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Interval()


class TimeField(FieldFactory, datetime.time):
    """Representation of a time field"""

    field_type = datetime.time

    def __new__(cls, with_timezone: bool = False, **kwargs: Any) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Time(kwargs.get("with_timezone") or False)


class JSONField(FieldFactory, pydantic.Json):
    """Representation of a JSONField"""

    field_type = Any

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.JSON()

    @classmethod
    def get_default_value(cls, field_obj: BaseFieldType, original_fn: Any = None) -> Any:
        default = original_fn()
        # copy mutable structures
        if isinstance(default, list | dict):
            default = copy.deepcopy(default)
        return default

    @classmethod
    def customize_default_for_server_default(
        cls, field_obj: BaseFieldType, default: Any, original_fn: Any = None
    ) -> Any:
        if callable(default):
            default = default()
        if not isinstance(default, str):
            default = orjson.dumps(default)
        return sqlalchemy.text(":value").bindparams(value=default)


class BinaryField(FieldFactory, bytes):
    """Representation of a binary"""

    field_type = bytes

    def __new__(cls, *, max_length: int | None = None, **kwargs: Any) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.LargeBinary(length=kwargs.get("max_length"))


class UUIDField(FieldFactory, uuid.UUID):
    """Representation of a uuid"""

    field_type = uuid.UUID

    def __new__(cls, **kwargs: Any) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        return sqlalchemy.Uuid(as_uuid=True, native_uuid=True)


class ChoiceField(FieldFactory):
    """Representation of an Enum"""

    def __new__(
        cls,
        choices: type[enum.Enum],
        **kwargs: Any,
    ) -> BaseFieldType:
        return super().__new__(cls, choices=choices, **kwargs)

    @classmethod
    def get_pydantic_type(cls, kwargs: dict[str, Any]) -> Any:
        """Returns the type for pydantic"""
        return kwargs["choices"]

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        choice_class = kwargs.get("choices")
        if choice_class is None or not isinstance(choice_class, EnumMeta):
            raise FieldDefinitionError("ChoiceField choices must be an Enum")

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> sqlalchemy.Enum:
        choice_class = kwargs.get("choices")
        return sqlalchemy.Enum(choice_class)

    @classmethod
    def customize_default_for_server_default(
        cls, field_obj: BaseFieldType, default: Any, original_fn: Any = None
    ) -> Any:
        if callable(default):
            default = default()
        return sqlalchemy.text(":value").bindparams(value=default.name)


class CharChoiceField(ChoiceField):
    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        kwargs.setdefault("max_length", 30)
        # we need to rename max_length. Pydantic will raise otherwise a TypeError
        max_length = kwargs["key_max_length"] = kwargs.pop("max_length")
        choice_class = kwargs.get("choices")
        if max_length is not None:
            for k in choice_class.__members__:
                if len(k) > max_length:
                    raise FieldDefinitionError(
                        f"ChoiceField choice name {k} is longer than {max_length} characters."
                        "Alternatively raise the max_length via explicitly passing a max_length argument or "
                        "set it to None (less performant)."
                    )

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> Any:
        max_length: int | None = kwargs.get("key_max_length")
        return (
            sqlalchemy.Text(collation=kwargs.get("collation"))
            if max_length is None
            else sqlalchemy.String(length=max_length, collation=kwargs.get("collation"))
        )

    @classmethod
    def to_model(
        cls, field_obj: BaseFieldType, field_name: str, value: Any, original_fn: Any = None
    ) -> dict[str, Any]:
        return {field_name: cls.check(field_obj, value=value)}

    @classmethod
    def clean(
        cls,
        field_obj: BaseFieldType,
        field_name: str,
        value: Any,
        for_query: bool = False,
        original_fn: Any = None,
    ) -> dict[str, Any]:
        return {field_name: cls.check(field_obj, value=value).name}

    @classmethod
    def check(cls, field_obj: BaseFieldType, value: Any, original_fn: Any = None) -> Any:
        if isinstance(value, Enum):
            value = value.name
        if not isinstance(value, str):
            raise ValueError("Value must be a string or enum member.")
        try:
            return field_obj.choices.__members__[value]
        except KeyError:
            raise ValueError(f"Invalid enum key {value}.") from None

    @classmethod
    def get_default_value(cls, field_obj: BaseFieldType, original_fn: Any = None) -> Any:
        default = original_fn()
        return default.name


class PasswordField(CharField):
    """
    Representation of a Password
    """

    def __new__(
        cls,
        derive_fn: Callable[[str], str] | None = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs.setdefault("keep_original", derive_fn is not None)
        return super().__new__(cls, derive_fn=derive_fn, **kwargs)

    @classmethod
    def get_embedded_fields(
        cls,
        field_obj: BaseFieldType,
        name: str,
        fields: dict[str, "BaseFieldType"],
        original_fn: Any = None,
    ) -> dict[str, "BaseFieldType"]:
        retdict: dict[str, BaseFieldType] = {}
        # TODO: check if it works in embedded settings or embed_field is required
        if field_obj.keep_original:
            original_pw_name = f"{name}_original"
            if original_pw_name not in fields:
                retdict[original_pw_name] = cast(
                    BaseFieldType,
                    _PlaceholderField(
                        pydantic_field_type=str,
                        null=True,
                        exclude=True,
                        read_only=True,
                        name=original_pw_name,
                        owner=field_obj.owner,
                    ),
                )

        return retdict

    @classmethod
    def to_model(
        cls, field_obj: BaseFieldType, field_name: str, value: Any, original_fn: Any = None
    ) -> dict[str, Any]:
        if isinstance(value, tuple | list):
            # despite an != should not be a problem here, make sure that strange logics
            # doesn't leak timings of the password
            if not compare_digest(value[0], value[1]):
                raise ValueError("Password doesn't match.")
            else:
                value = value[0]
        retval: dict[str, Any] = {}
        phase = CURRENT_PHASE.get()
        derive_fn = cast(Callable[[str], str] | None, field_obj.derive_fn)
        if phase in {"set", "init"} and derive_fn is not None:
            retval[field_name] = derive_fn(value)
            if getattr(field_obj, "keep_original", False):
                retval[f"{field_name}_original"] = value
        else:
            retval[field_name] = value
            # blank after saving or loading
            if phase in {"post_insert", "post_update", "load"} and getattr(
                field_obj, "keep_original", False
            ):
                retval[f"{field_name}_original"] = None

        return retval

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("secret", True)
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class EmailField(CharField):
    field_type = EmailStr

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


UrlString = Annotated[AnyUrl, pydantic.AfterValidator(lambda v: v if v is None else str(v))]


class URLField(CharField):
    field_type = UrlString  # type: ignore

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class IPAddressField(FieldFactory, str):
    field_type = IPvAnyAddress

    def __new__(
        cls,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, kwargs: dict[str, Any]) -> IPAddress:
        return IPAddress()

    @staticmethod
    def is_native_type(value: str) -> bool:
        return isinstance(value, ipaddress.IPv4Address | ipaddress.IPv6Address)

    # overwrite
    @classmethod
    def check(cls, field_obj: BaseFieldType, value: Any, original_fn: Any = None) -> Any:
        if cls.is_native_type(value):
            return value

        try:
            return ipaddress.ip_address(value)
        except ValueError:
            raise ValueError("Must be a real IP.")  # noqa


Monkay(
    globals(),
    deprecated_lazy_imports={
        "ComputedField": {
            "path": ".computed_field.ComputedField",
            "reason": "The import path changed.",
            "new_attribute": "edgy.core.db.fields.ComputedField",
        },
        "PlaceholderField": {
            "path": lambda: _PlaceholderField,
            "reason": "The import path changed.",
            "new_attribute": "edgy.core.db.fields.PlaceholderField",
        },
        "AutoNowMixin": {
            "path": lambda: _AutoNowMixin,
            "reason": "We export mixins now from edgy.core.db.fields.mixins.",
            "new_attribute": "edgy.core.db.fields.mixins.AutoNowMixin",
        },
    },
)



import os

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class AdminConfig(BaseSettings):
    model_config = SettingsConfigDict(extra="allow", arbitrary_types_allowed=True)
    admin_prefix_url: str | None = None
    admin_extra_templates: list[str | os.PathLike] = Field(default_factory=list)
    title: str = "Edgy Admin"
    menu_title: str = "Edgy Admin"
    favicon: str = "https://raw.githubusercontent.com/dymmond/edgy/refs/heads/main/docs/statics/images/favicon.ico"
    sidebar_bg_colour: str = "#1C4C74"
    dashboard_title: str = "Edgy Admin Dashboard"
    SECRET_KEY: str | bytes = Field(default_factory=lambda: os.urandom(64))



from __future__ import annotations

import sys
from collections.abc import AsyncGenerator, Hashable, Iterable
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from pydantic import BaseModel

if TYPE_CHECKING:
    from edgy.core.db.querysets import QuerySet

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self


class BasePage(BaseModel):
    content: list[Any]
    is_first: bool
    is_last: bool


class Page(BasePage):
    current_page: int = 1
    next_page: int | None
    previous_page: int | None


PageType = TypeVar("PageType", bound=BasePage)


class BasePaginator(Generic[PageType]):
    order_by: tuple[str, ...]

    def __init__(
        self,
        queryset: QuerySet,
        page_size: int,
        next_item_attr: str = "",
        previous_item_attr: str = "",
    ) -> None:
        self._reverse_paginator: Self | None = None
        self.page_size = int(page_size)
        if page_size < 0:
            raise ValueError("page_size must be at least 0")
        if len(queryset._order_by) == 0:
            raise ValueError("You must pass a QuerySet with .order_by(*criteria)")
        self.next_item_attr = next_item_attr
        self.previous_item_attr = previous_item_attr
        self.queryset = queryset
        if self.previous_item_attr or self.next_item_attr:
            # copy for not leaking attributes in the cache
            old_queryset = self.queryset
            self.queryset = self.queryset.all()
            # but keep some caches
            self.queryset._cache_count = old_queryset._cache_count
            self.queryset._cached_select_related_expression = (
                old_queryset._cached_select_related_expression
            )
            self.queryset._cached_select_with_tables = old_queryset._cached_select_with_tables
        self.order_by = queryset._order_by
        self._page_cache: dict[Hashable, PageType] = {}

    def clear_caches(self) -> None:
        self._page_cache.clear()
        self.queryset.all(clear_cache=True)
        if self._reverse_paginator:
            self._reverse_paginator = None

    async def get_amount_pages(self) -> int:
        if not self.page_size:
            return 1
        count, remainder = divmod(await self.get_total(), self.page_size)
        return count + (1 if remainder else 0)

    async def get_total(self) -> int:
        return await self.queryset.count()

    async def get_page(self) -> PageType:
        raise NotImplementedError()

    def shall_drop_first(self, is_first: bool) -> bool:
        return bool(self.next_item_attr and not is_first)

    def convert_to_page(self, inp: Iterable, /, is_first: bool, reverse: bool = False) -> PageType:
        last_item: Any = None
        drop_first = self.shall_drop_first(is_first)
        result_list = []
        item_counter = 0
        for item in inp:
            if reverse:
                if self.next_item_attr:
                    setattr(item, self.next_item_attr, last_item)
                if last_item is not None and self.previous_item_attr:
                    setattr(last_item, self.previous_item_attr, item)
            else:
                if self.previous_item_attr:
                    setattr(item, self.previous_item_attr, last_item)
                if last_item is not None and self.next_item_attr:
                    setattr(last_item, self.next_item_attr, item)
            if (not drop_first and item_counter >= 1) or (drop_first and item_counter >= 2):
                result_list.append(last_item)
            last_item = item
            item_counter += 1
        min_size = self.page_size + 1
        if self.previous_item_attr and not is_first:
            min_size += 1
        is_last = bool(self.page_size == 0 or item_counter < min_size)
        if is_last and (
            (not drop_first and item_counter >= 1) or (drop_first and item_counter >= 2)
        ):
            if reverse and self.previous_item_attr:
                setattr(last_item, self.previous_item_attr, None)
            elif not reverse and self.next_item_attr:
                setattr(last_item, self.next_item_attr, None)
            result_list.append(last_item)
        if reverse:
            result_list.reverse()
            return cast(
                PageType, BasePage(content=result_list, is_first=is_last, is_last=is_first)
            )
        else:
            return cast(
                PageType, BasePage(content=result_list, is_first=is_first, is_last=is_last)
            )

    async def paginate_queryset(
        self, queryset: QuerySet, is_first: bool = True, prefill: Iterable | None = None
    ) -> AsyncGenerator[BasePage, None]:
        container: list = []
        if prefill is not None:
            container.extend(prefill)
        page: PageType | None = None
        min_size = self.page_size + 1
        if self.previous_item_attr and not is_first:
            min_size += 1
        if queryset.database.force_rollback:
            for item in await queryset:
                container.append(item)
                if self.page_size and len(container) >= min_size:
                    page = BasePaginator.convert_to_page(self, container, is_first=is_first)
                    yield page
                    if self.previous_item_attr and is_first:
                        min_size += 1
                    is_first = False
                    container = [page.content[-1], item] if self.previous_item_attr else [item]
        else:
            async for item in queryset:
                container.append(item)
                if self.page_size and len(container) >= min_size:
                    page = BasePaginator.convert_to_page(self, container, is_first=is_first)
                    yield page
                    if self.previous_item_attr and is_first:
                        min_size += 1
                    is_first = False
                    container = [page.content[-1], item] if self.previous_item_attr else [item]
        if page is None or not page.is_last:
            yield BasePaginator.convert_to_page(self, container, is_first=is_first)

    async def paginate(self) -> AsyncGenerator[PageType, None]:
        raise NotImplementedError()

    def get_reverse_paginator(self) -> Self:
        if self._reverse_paginator is None:
            self._reverse_paginator = type(self)(
                self.queryset.reverse(),
                page_size=self.page_size,
                next_item_attr=self.next_item_attr,
                previous_item_attr=self.previous_item_attr,
            )
            self._reverse_paginator._reverse_paginator = self
        return self._reverse_paginator


class Paginator(BasePaginator[Page]):
    async def paginate(self, start_page: int = 1) -> AsyncGenerator[Page, None]:
        query = self.queryset
        if start_page > 1:
            offset = self.page_size * (start_page - 1)
            if self.previous_item_attr:
                offset = max(offset - 1, 0)
            if offset > 0:
                query = query.offset(offset)
        counter = 1
        async for page_obj in self.paginate_queryset(query, is_first=start_page == 1):
            yield Page(
                **page_obj.__dict__,
                next_page=None if page_obj.is_last else counter + 1,
                previous_page=None if page_obj.is_first else counter - 1,
                current_page=counter,
            )
            counter += 1

    def convert_to_page(
        self, inp: Iterable, /, page: int, is_first: bool, reverse: bool = False
    ) -> Page:
        page_obj: BasePage = super().convert_to_page(
            inp,
            is_first=is_first,
            reverse=reverse,
        )
        return Page(
            **page_obj.__dict__,
            current_page=page,
            next_page=None if page_obj.is_last else page + 1,
            previous_page=None if page_obj.is_first else page + 1,
        )

    async def _get_page(self, page: int, reverse: bool = False) -> Page:
        # this is a special intern reverse which really reverses a page
        if page < 0 and self.page_size:
            return await self.get_reverse_paginator()._get_page(-page, reverse=True)
        else:
            offset = self.page_size * (page - 1)
        if self.previous_item_attr:
            offset = max(offset - 1, 0)
        if self.queryset._cache_fetch_all:
            resultarr = await self.queryset
            if self.page_size:
                resultarr = resultarr[offset : offset + self.page_size + 1]
            return self.convert_to_page(
                resultarr, page=page, is_first=offset == 0, reverse=reverse
            )
        query = self.queryset
        query = query.offset(offset)
        if self.page_size:
            query = query.limit(self.page_size + 1)
        return self.convert_to_page(await query, page=page, is_first=offset == 0, reverse=reverse)

    async def get_page(self, page: int = 1) -> Page:
        if page == 0 or not isinstance(page, int):
            raise ValueError(f"Invalid page parameter value: {page!r}")
        if self.page_size == 0:
            # for caching, there are no other pages, this does not apply to cursor!
            page = 1
        if page in self._page_cache:
            return self._page_cache[page]
        page_obj = await self._get_page(page=page)

        self._page_cache[page] = page_obj
        return page_obj



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

import anyio
import orjson
from lilya.controllers import Controller
from lilya.exceptions import NotFound  # noqa
from lilya.requests import Request
from lilya.responses import JSONResponse, RedirectResponse
from lilya.templating.controllers import TemplateController

import edgy
from edgy.contrib.admin.mixins import AdminMixin
from edgy.contrib.admin.utils.messages import add_message
from edgy.contrib.pagination import Paginator
from edgy.core.db.fields.file_field import ConcreteFileField
from edgy.core.db.fields.many_to_many import BaseManyToManyForeignKeyField
from edgy.core.db.relationships.related_field import RelatedField
from edgy.exceptions import ObjectNotFound

from .utils.models import (
    add_to_recent_models,
    get_model_json_schema,
    get_recent_models,
    get_registered_models,
)
from .utils.models import get_model as _get_model

if TYPE_CHECKING:
    from edgy.core.db.models.model import Model
    from edgy.core.db.querysets.base import QuerySet


def get_registered_model(model: str) -> type[Model]:
    try:
        return _get_model(model)
    except LookupError:
        raise NotFound() from None


class JSONSchemaView(Controller):
    def get(self, request: Request) -> JSONResponse:
        phase = request.query_params.get("phase", "view")
        with_defaults = request.query_params.get("cdefaults") == "true"
        model_name = request.path_params.get("name")
        reftemplate = "../{model}/json"
        reftemplate = f"{reftemplate}?phase={phase}"
        if with_defaults:
            reftemplate = f"{reftemplate}&cdefaults=true"
        try:
            with JSONResponse.with_transform_kwargs({"json_encode_fn": orjson.dumps}):
                return JSONResponse(
                    get_model_json_schema(
                        model_name,
                        include_callable_defaults=with_defaults,
                        ref_template=reftemplate,
                        no_check_admin_models=True,
                        phase=phase,
                    )
                )
        except LookupError:
            raise NotFound() from None


class AdminDashboard(AdminMixin, TemplateController):
    """
    View for the administration dashboard page.
    """

    template_name = "admin/dashboard.html"

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        """
        Prepares the context data for the dashboard template.

        Args:
            request: The incoming Lilya Request object.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary containing context data, including the page title.
        """
        context = await super().get_context_data(request, **kwargs)

        models = get_registered_models()
        model_stats: list[dict] = []

        async with anyio.create_task_group() as tg:
            for name, model in models.items():
                tg.start_soon(self._add_model_stat, model_stats, name, model)  # type: ignore

        total_records = sum(m["count"] for m in model_stats)
        top_model = max(
            model_stats, key=lambda m: m["count"], default={"verbose": "N/A", "count": 0}
        )

        context.update(
            {
                "title": "Dashboard",
                "models": sorted(model_stats, key=lambda m: m["verbose"]),
                "total_records": total_records,
                "top_model": top_model,
                "recent_models": get_recent_models(),
            }
        )
        return context

    async def _add_model_stat(self, model_stats: list, name: str, model: edgy.Model) -> None:
        try:
            count = await model.query.count()
        except Exception:
            count = 0

        model_stats.append(
            {
                "name": name,
                "verbose": model.__name__,
                "count": count,
                "no_admin_create": model.meta.no_admin_create,
            }
        )

    async def get(self, request: Request) -> Any:
        """
        Handles GET requests for the administration dashboard page.

        Args:
            request: The incoming Starlette Request object.

        Returns:
            The rendered template response for the dashboard.
        """
        return await self.render_template(request)


class ModelOverview(AdminMixin, TemplateController):
    """
    View for listing all registered models in the admin interface.
    """

    template_name = "admin/models.html"

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        """
        Prepares the context data for the models list template.

        Retrieves all registered models and adds them to the context.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary containing context data, including the page title
            and a list of registered models.
        """
        context = await super().get_context_data(request, **kwargs)
        context.update({"title": "Models", "models": get_registered_models()})
        return context

    async def get(self, request: Request) -> Any:
        """
        Handles GET requests for the registered models list page.

        Args:
            request: The incoming Starlette Request object.

        Returns:
            The rendered template response for the models list.
        """
        return await self.render_template(request)


class ModelDetailView(AdminMixin, TemplateController):
    """
    View for displaying details of a specific model, typically listing its objects.
    """

    template_name = "admin/model_detail.html"

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        """
        Prepares the context data for the model detail template.

        Retrieves the model name from the request, finds the corresponding
        model class, fetches a limited number of objects for that model,
        and adds them to the context.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A dictionary containing context data, including the model name,
            the model class, and a list of objects belonging to the model.

        Raises:
            NotFound: If the model name is not found in the registered models.
        """
        context = await super().get_context_data(request, **kwargs)
        model_name = request.path_params.get("name")

        # For the search
        query = request.query_params.get("q", "").strip()
        page = int(request.query_params.get("page", 1))
        page_size = int(request.query_params.get("per_page", 25))
        page_size = min(max(page_size, 1), 250)

        model = get_registered_model(model_name)
        marshall_class = model.get_admin_marshall_class(phase="list", for_schema=False)
        add_to_recent_models(model)

        queryset = model.query.all()

        if query:
            filters = []
            for field in model.model_fields.values():
                if field.annotation is str:
                    column = model.table.c.get(field.name)
                    if column is not None:
                        filters.append(column.ilike(f"%{query}%"))

            if filters:
                queryset = queryset.filter(*filters)
        paginator = Paginator(queryset.order_by(*model.pknames), page_size=page_size)
        page_obj = await paginator.get_page(page)

        total_pages = await paginator.get_amount_pages()

        context.update(
            {
                "title": f"{model.__name__} Details",
                "model": model,
                "marshall_class": marshall_class,
                "page": page_obj,
                "model_name": model_name,
                "query": query,
                "per_page": page_size,
                "total_pages": total_pages,
            }
        )
        return context

    async def get(self, request: Request, **kwargs: Any) -> Any:
        """
        Handles GET requests for the model detail page.

        Renders the template configured for this view, populated with
        context data prepared by `get_context_data`.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            The rendered template response for the model detail page.
        """
        return await self.render_template(request, **kwargs)

    async def post(self, request: Request, **kwargs: Any) -> Any:
        """
        Handles POST requests for the page.

        Currently, this method just re-renders the template, implying
        that saving/processing the POST data is handled elsewhere or not implemented here.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            The rendered template response for the page.
        """
        return await self.render_template(request, **kwargs)


class BaseObjectView:
    """
    Base class for admin views that operate on a single model object.
    Provides utility methods for extracting and parsing object IDs.
    """

    async def save_model(
        self, instance: edgy.Model | type[edgy.Model], json_data: dict, create: bool = False
    ) -> edgy.Model:
        """
        Saves an Edgy model instance based on form data.

        This function updates the fields of a given Edgy model instance
        using data from a FormData object. It specifically handles foreign key
        fields by fetching the related model instance before updating.
        Finally, it saves the updated model instance.

        Args:
            instance: The Edgy model instance to be updated and saved.
                      Note: The type hint 'type[edgy.Model]' might be intended
                      as the instance itself, not the class type.
            json_data: A json dict to update the model.
            create: Is it an update or creation.
        """
        # retrieve fields for extraction, should be broader then get_admin_marshall_for_save
        marshall_class = instance.get_admin_marshall_class(
            phase="create" if create else "update", for_schema=False
        )
        model_fields = marshall_class.model_fields
        data: dict = {}

        for key, value in json_data.items():
            field = model_fields.get(key)

            if field is None or getattr(field, "read_only", False):
                continue
            data[key] = value

        if create:
            marshall = instance.get_admin_marshall_for_save(**data)
        else:
            marshall = instance.get_admin_marshall_for_save(cast("edgy.Model", instance), **data)
        return (await marshall.save()).instance

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        context: dict = await super().get_context_data(request, **kwargs)

        model_name = request.path_params.get("name")

        model = get_registered_model(model_name)
        add_to_recent_models(model)

        context.update(
            {
                "model": model,
                "model_name": model_name,
            }
        )
        return context


class ModelObjectDetailView(BaseObjectView, AdminMixin, TemplateController):
    """
    View for displaying the details of a single object of a specific model.
    """

    template_name = "admin/model_object_detail.html"

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        """
        Prepares the context data for the model object detail template.

        Retrieves the model name and object ID from the request, finds the
        corresponding model and fetches the specific instance by its ID.
        Adds the instance, model information, and title to the context.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A dictionary containing context data, including the page title,
            the model instance, the model class, and the model name.

        Raises:
            NotFound: If the model name or the specific instance is not found.
        """
        context = await super().get_context_data(request, **kwargs)  # noqa
        model: type[Model] = context["model"]

        instance = await model.query.get_or_none(pk=self.get_object_pk(request))
        if not instance:
            raise NotFound()
        marshall_class = instance.get_admin_marshall_class(phase="view", for_schema=False)
        marshall = marshall_class(instance=instance)
        relationship_fields = {}
        overwrite_values = {}

        for name, field in marshall_class.model_fields.items():
            if isinstance(field, BaseManyToManyForeignKeyField):
                relationship_fields[name] = "many_to_many"
                overwrite_values[name] = await getattr(instance, name).all()
            elif isinstance(field, RelatedField):
                relationship_fields[name] = "related_field"
                overwrite_values[name] = await getattr(instance, name).all()
            elif name in model.meta.foreign_key_fields:
                relationship_fields[name] = "foreign_key"
                overwrite_values[name] = getattr(instance, name)
            elif isinstance(field, ConcreteFileField):
                overwrite_values[name] = getattr(instance, name)

        values = marshall.model_dump(exclude=overwrite_values.keys())
        values.update(overwrite_values)

        context.update(
            {
                "title": f"{model.__name__.capitalize()} #{instance}",
                "marshall_class": marshall_class,
                "values": values,
                "object_pk": self.create_object_pk(instance),
                "relationship_fields": relationship_fields,
            }
        )
        return context

    async def get(self, request: Request, **kwargs: Any) -> Any:
        """
        Handles GET requests for the model object detail page.

        Renders the template configured for this view, populated with
        context data prepared by `get_context_data`.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            The rendered template response for the object detail page.
        """
        return await self.render_template(request, **kwargs)

    async def post(self, request: Request, **kwargs: Any) -> Any:
        """
        Handles POST requests for the model object detail page.

        Currently, this method just re-renders the template, implying
        that saving/editing is handled elsewhere or not implemented here.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            The rendered template response for the object detail page.
        """
        return await self.render_template(request, **kwargs)


class ModelObjectEditView(BaseObjectView, AdminMixin, TemplateController):
    template_name = "admin/model_object_edit.html"

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        """
        Prepares the context data for the model edit template.

        Retrieves the model name and object ID from the request, fetches
        the corresponding model and instance, and populates the context
        dictionary with information needed for rendering the edit form.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A dictionary containing context data for the template, including
            the page title, the model instance, the model class, and the model name.

        Raises:
            NotFound: If the model name or the specific instance is not found.
        """
        context = await super().get_context_data(request, **kwargs)  # noqa
        model = cast("type[Model]", context["model"])
        instance: Model | None = await cast("QuerySet", model.query).get_or_none(
            pk=self.get_object_pk(request)
        )
        if not instance:
            raise NotFound()
        context["title"] = f"Edit {instance}"
        context["object"] = instance
        marshall = instance.get_admin_marshall_class(phase="view", for_schema=False)(instance)
        json_values = marshall.model_dump_json(exclude_none=True)
        context["values_as_json"] = json_values
        context["object_pk"] = self.create_object_pk(instance)
        return context

    async def get(self, request: Request, **kwargs: Any) -> Any:
        """
        Handles GET requests to display the model edit form.

        Renders the template configured for this view, populated with
        context data prepared by `get_context_data`.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            The rendered template response containing the edit form.
        """
        return await self.render_template(request, **kwargs)

    async def post(self, request: Request, **kwargs: Any) -> RedirectResponse:
        """
        Handles POST requests to save the changes to a model instance.

        Retrieves the model name, object ID, and form data from the request,
        finds the corresponding model and instance, updates the instance
        using the form data via `save_model`, and redirects the user back
        to the edited object's detail page.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A RedirectResponse to the edited object's detail page after saving.

        Raises:
            NotFound: If the model name or the specific instance is not found.
        """
        model_name = request.path_params.get("name")
        obj_id = request.path_params.get("id")

        model = get_registered_model(model_name)

        try:
            instance: Model = await cast("QuerySet", model.query).get(
                pk=self.get_object_pk(request)
            )
        except ObjectNotFound:
            add_message("error", f"Model {model_name} with ID {obj_id} not found.")
            return RedirectResponse(f"{self.get_admin_prefix_url(request)}/models/{model_name}")

        await self.save_model(instance, orjson.loads((await request.form())["editor_data"]))

        add_message(
            "success",
            f"{instance} has been updated successfully.",
        )
        return RedirectResponse(
            f"{self.get_admin_prefix_url(request)}/models/{model_name}/{obj_id}"
        )


class ModelObjectDeleteView(AdminMixin, Controller):
    async def post(self, request: Request, **kwargs: Any) -> RedirectResponse:
        """
        Handles POST requests to delete a model instance.

        Retrieves the model name and object ID from the request's path parameters,
        finds the corresponding model and instance using the object ID,
        deletes the instance from the database, and redirects the user back
        to the model's listing page.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A RedirectResponse to the model's listing page after deletion.

        Raises:
            NotFound: If the model name or the specific instance is not found.
        """
        model_name = request.path_params.get("name")
        obj_id = request.path_params.get("id")

        model = get_registered_model(model_name)

        try:
            instance = await model.query.get(pk=self.get_object_pk(request))
        except ObjectNotFound:
            add_message("error", f"There is no record with this ID: '{obj_id}'.")
            return RedirectResponse(f"{self.get_admin_prefix_url(request)}/models/{model_name}")
        instance_name = str(instance)
        await instance.delete()

        add_message(
            "success",
            f"{model_name.capitalize()} #{instance_name} has been deleted successfully.",
        )
        return RedirectResponse(f"{self.get_admin_prefix_url(request)}/models/{model_name}")


class ModelObjectCreateView(BaseObjectView, AdminMixin, TemplateController):
    """
    View for displaying and processing the form to create a new model instance.
    """

    template_name = "admin/model_object_create.html"

    async def get_context_data(self, request: Request, **kwargs: Any) -> dict:
        """
        Prepares the context data for the model creation template.

        Retrieves the model name from the request, finds the corresponding
        model class, and adds it to the context.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A dictionary containing context data, including the page title,
            the model class, and the model name.

        Raises:
            NotFound: If the model name is not found in the registered models.
        """
        context = await super().get_context_data(request, **kwargs)
        model_name: str = context["model_name"]

        context.update(
            {
                "title": f"Create {model_name.capitalize()}",
            }
        )
        return context

    async def get(self, request: Request, **kwargs: Any) -> Any:
        """
        Handles GET requests to display the model creation form.

        Renders the template configured for this view, populated with
        context data prepared by `get_context_data`.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            The rendered template response containing the creation form.
        """
        model_name = request.path_params.get("name")

        model = get_registered_model(model_name)
        if (
            model.meta.no_admin_create
            or model.get_admin_marshall_class(
                phase="create", for_schema=False
            ).__incomplete_fields__
        ):
            add_message(
                "error",
                f"For {model.__name__.capitalize()} we cannot create a new instance.",
            )
            return RedirectResponse(
                f"{self.get_admin_prefix_url(request)}/models/{model.__name__}"
            )

        return await self.render_template(request, **kwargs)

    async def post(self, request: Request, **kwargs: Any) -> RedirectResponse:
        """
        Handles POST requests to create a new model instance.

        Retrieves the model name and form data from the request, finds the
        corresponding model class, creates a new instance from the form data,
        saves it to the database, and redirects to the detail page of the
        newly created object.

        Args:
            request: The incoming Starlette Request object.
            **kwargs: Additional keyword arguments from the path parameters.

        Returns:
            A RedirectResponse to the detail page of the newly created object.

        Raises:
            NotFound: If the model name is not found in the registered models.
        """
        model_name = request.path_params.get("name")

        model = get_registered_model(model_name)
        if (
            model.meta.no_admin_create
            or model.get_admin_marshall_class(
                phase="create", for_schema=False
            ).__incomplete_fields__
        ):
            add_message(
                "error",
                f"For {model.__name__.capitalize()} we cannot create a new instance.",
            )
            return RedirectResponse(
                f"{self.get_admin_prefix_url(request)}/models/{model.__name__}"
            )

        instance = await self.save_model(
            model, orjson.loads((await request.form())["editor_data"]), create=True
        )
        obj_id = self.create_object_pk(instance)
        add_message(
            "success",
            f"{model_name.capitalize()} #{instance} has been created successfully.",
        )
        return RedirectResponse(
            f"{self.get_admin_prefix_url(request)}/models/{model_name}/{obj_id}"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from pydantic import ConfigDict

from edgy.core.marshalls.base import Marshall
from edgy.core.marshalls.config import ConfigMarshall

if TYPE_CHECKING:
    from edgy.core.db.models import Model


class AdminMixin:
    """
    Admin related mixin.
    """

    @classmethod
    def get_admin_marshall_config(
        cls: type[Model], *, phase: str, for_schema: bool
    ) -> dict[str, Any]:
        """
        Shortcut for updating the marshall_config of the admin marshall.

        Can be dynamic for the current user.
        """
        return {"fields": ["__all__"]}

    @classmethod
    def get_admin_marshall_class(
        cls: type[Model], *, phase: str, for_schema: bool = False
    ) -> type[Marshall]:
        """
        Generate a marshall class for the admin.

        Can be dynamic for the current user.
        """

        class AdminMarshall(Marshall):
            # forbid triggers additionalProperties=false
            model_config: ClassVar[ConfigDict] = ConfigDict(
                title=cls.__name__, extra="forbid" if for_schema else None
            )
            marshall_config = ConfigMarshall(
                model=cls,
                **cls.get_admin_marshall_config(phase=phase, for_schema=for_schema),  # type: ignore
            )

        return AdminMarshall

    @classmethod
    def get_admin_marshall_for_save(
        cls: type[Model], instance: Model | None = None, /, **kwargs: Any
    ) -> Marshall:
        """Generate a marshall instance from an instance for the admin.

        Can be dynamic for the current user. Called in the create/edit path.
        """
        return cls.get_admin_marshall_class(
            phase="update" if instance is not None else "create", for_schema=False
        )(instance, **kwargs)



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

import orjson
from pydantic import BaseModel

from edgy.core.db.context_vars import (
    MODEL_GETATTR_BEHAVIOR,
)

if TYPE_CHECKING:
    from edgy.core.db.models.metaclasses import MetaInfo

_empty = cast(set[str], frozenset())


class DumpMixin:
    """
    Dump mixin.
    """

    meta: ClassVar[MetaInfo]

    def model_dump(self: BaseModel, show_pk: bool | None = None, **kwargs: Any) -> dict[str, Any]:
        """
        An updated version of the model dump.
        It can show the pk always and handles the exclude attribute on fields correctly and
        contains the custom logic for fields with getters

        Extra Args:
            show_pk: bool - Enforces showing the primary key in the model_dump.
        """
        meta = self.meta
        # we want a copy
        _exclude: set[str] | dict[str, Any] | None = kwargs.pop("exclude", None)
        if _exclude is None:
            initial_full_field_exclude: set[str] = _empty
            # must be a writable dict
            exclude_passed: dict[str, Any] = {}
            exclude_second_pass: dict[str, Any] = {}
        elif isinstance(_exclude, dict):
            initial_full_field_exclude = {k for k, v in _exclude.items() if v is True}
            exclude_passed = copy.copy(_exclude)
            exclude_second_pass = _exclude
        else:
            initial_full_field_exclude = set(_exclude)
            exclude_passed = dict.fromkeys(initial_full_field_exclude, True)
            exclude_second_pass = exclude_passed.copy()

        need_second_pass: set[str] = set()
        for field_name in meta.special_getter_fields:
            exclude_passed[field_name] = True
            if (
                field_name not in initial_full_field_exclude
                and not meta.fields[field_name].exclude
            ):
                need_second_pass.add(field_name)
        for field_name in meta.foreign_key_fields:
            field = meta.fields[field_name]
            # when excluded we don't need to consider to add a second pass.
            if field_name in initial_full_field_exclude or field.exclude:
                continue
            if field.target.meta.needs_special_serialization:
                exclude_passed[field_name] = True
                need_second_pass.add(field_name)
        include: set[str] | dict[str, Any] | None = kwargs.pop("include", None)
        mode: Literal["json", "python"] = kwargs.pop("mode", "python")

        should_show_pk = self.__show_pk__ if show_pk is None else show_pk
        result_dict: dict[str, Any] = super().model_dump(
            exclude=exclude_passed, include=include, mode=mode, **kwargs
        )
        token = MODEL_GETATTR_BEHAVIOR.set("passdown")
        try:
            for field_name in need_second_pass:
                # include or pkname (if should_show_pk)
                if not (
                    (should_show_pk and field_name in self.pknames)
                    or include is None
                    or field_name in include
                ):
                    continue
                field = meta.fields[field_name]
                try:
                    retval = getattr(self, field_name)
                except AttributeError:
                    continue
                sub_include = None
                if isinstance(include, dict):
                    sub_include = include.get(field_name, None)
                    if sub_include is True:
                        sub_include = None
                sub_exclude = exclude_second_pass.get(field_name, None)
                assert sub_exclude is not True, "field should have been excluded"
                # for e.g. CompositeFields which generate via __get__
                if isinstance(retval, BaseModel):
                    retval = retval.model_dump(
                        include=sub_include, exclude=sub_exclude, mode=mode, **kwargs
                    )
                else:
                    assert sub_include is None, "sub include filters for no pydantic model"
                    assert sub_exclude is None, "sub exclude filters for no pydantic model"
                    if mode == "json" and not getattr(field, "unsafe_json_serialization", False):
                        # skip field if it isn't a BaseModel and the mode is json and
                        # unsafe_json_serialization is not set
                        # currently unsafe_json_serialization exists only on CompositeFields
                        continue
                alias: str = field_name
                if getattr(field, "serialization_alias", None):
                    alias = cast(str, field.serialization_alias)
                elif getattr(field, "alias", None):
                    alias = field.alias
                result_dict[alias] = retval
        finally:
            MODEL_GETATTR_BEHAVIOR.reset(token)
        return result_dict

    def model_dump_json(self, **kwargs: Any) -> str:
        return orjson.dumps(self.model_dump(mode="json", **kwargs)).decode()



from __future__ import annotations

from pathlib import Path
from typing import Any

from lilya.apps import ChildLilya, Lilya
from lilya.middleware import DefineMiddleware
from lilya.middleware.session_context import SessionContextMiddleware
from lilya.requests import Request
from lilya.routing import RoutePath
from lilya.templating import Jinja2Template

from edgy.contrib.admin.views import (
    AdminDashboard,
    JSONSchemaView,
    ModelDetailView,
    ModelObjectCreateView,
    ModelObjectDeleteView,
    ModelObjectDetailView,
    ModelObjectEditView,
    ModelOverview,
)

templates = Jinja2Template(directory=str(Path(__file__).resolve().parent / "templates"))
templates.env.globals["getattr"] = getattr


async def not_found(request: Request, exc: Exception) -> Any:
    return templates.get_template_response(
        request,
        "404.html",
        context={"title": "Not Found"},
    )


def create_admin_app(*, session_sub_path: str = "") -> Lilya:
    middleware = []
    if session_sub_path:
        middleware.append(DefineMiddleware(SessionContextMiddleware, sub_path=session_sub_path))
    else:
        middleware.append(DefineMiddleware(SessionContextMiddleware))
    return ChildLilya(
        routes=[
            RoutePath("/", handler=AdminDashboard, name="admin"),
            RoutePath("/models", handler=ModelOverview, name="models"),
            RoutePath("/models/{name}", handler=ModelDetailView, name="model-details"),
            RoutePath("/models/{name}/json", handler=JSONSchemaView, name="model-schema"),
            RoutePath("/models/{name}/create", handler=ModelObjectCreateView, name="model-create"),
            RoutePath("/models/{name}/{id}", handler=ModelObjectDetailView, name="model-object"),
            RoutePath(
                "/models/{name}/{id}/edit",
                handler=ModelObjectEditView,
                name="model-edit",
                methods=["GET", "POST"],
            ),
            RoutePath(
                "/models/{name}/{id}/delete",
                handler=ModelObjectDeleteView,
                name="model-delete",
                methods=["POST"],
            ),
        ],
        middleware=middleware,
        exception_handlers={404: not_found},
    )



from __future__ import annotations

import contextlib
import copy
import inspect
import sys
import warnings
from collections.abc import Awaitable, Callable, Collection, Sequence
from functools import partial
from itertools import chain
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

import sqlalchemy
from pydantic import BaseModel

from edgy.core.db.constants import CASCADE
from edgy.core.db.context_vars import (
    CURRENT_FIELD_CONTEXT,
    CURRENT_INSTANCE,
    CURRENT_MODEL_INSTANCE,
    EXPLICIT_SPECIFIED_VALUES,
    MODEL_GETATTR_BEHAVIOR,
    NO_GLOBAL_FIELD_CONSTRAINTS,
    get_schema,
)
from edgy.core.db.datastructures import Index, UniqueConstraint
from edgy.core.db.fields.base import BaseForeignKey
from edgy.core.db.fields.many_to_many import BaseManyToManyForeignKeyField
from edgy.core.db.models.metaclasses import MetaInfo
from edgy.core.db.models.types import BaseModelType
from edgy.core.db.models.utils import build_pkcolumns
from edgy.core.db.relationships.related_field import RelatedField
from edgy.core.utils.db import check_db_connection, hash_names
from edgy.core.utils.models import create_edgy_model
from edgy.exceptions import ForeignKeyBadConfigured, ModelCollisionError, ObjectNotFound
from edgy.types import Undefined

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self

if TYPE_CHECKING:
    from databasez.core.transaction import Transaction

    from edgy.core.connection.database import Database
    from edgy.core.connection.registry import Registry
    from edgy.core.db.fields.types import FIELD_CONTEXT_TYPE, BaseFieldType
    from edgy.core.db.models.model import Model
    from edgy.core.db.querysets.base import QuerySet


_empty = cast(set[str], frozenset())


class _EmptyClass: ...


_removed_copy_keys = {
    *BaseModel.__dict__.keys(),
    "_db_loaded",
    "_db_deleted",
    "_pkcolumns",
    "_table",
    "_db_schemas",
    "__proxy_model__",
    "meta",
    "transaction",
}
_removed_copy_keys.difference_update(
    {*_EmptyClass.__dict__.keys(), "__annotations__", "__module__"}
)


def _check_replace_related_field(
    replace_related_field: bool
    | type[BaseModelType]
    | tuple[type[BaseModelType], ...]
    | list[type[BaseModelType]],
    model: type[BaseModelType],
) -> bool:
    if isinstance(replace_related_field, bool):
        return replace_related_field
    if not isinstance(replace_related_field, tuple | list):
        replace_related_field = (replace_related_field,)
    return any(refmodel is model for refmodel in replace_related_field)


def _set_related_field(
    target: type[BaseModelType],
    *,
    foreign_key_name: str,
    related_name: str,
    source: type[BaseModelType],
    replace_related_field: bool
    | type[BaseModelType]
    | tuple[type[BaseModelType], ...]
    | list[type[BaseModelType]],
) -> None:
    if replace_related_field is not True and related_name in target.meta.fields:
        # is already correctly set, required for migrate of model_apps with registry set
        related_field = target.meta.fields[related_name]
        if (
            related_field.related_from is source
            and related_field.foreign_key_name == foreign_key_name
        ):
            return
        # required for copying
        if related_field.foreign_key_name != foreign_key_name or _check_replace_related_field(
            replace_related_field, related_field.related_from
        ):
            raise ForeignKeyBadConfigured(
                f"Multiple related_name with the same value '{related_name}' found to the same target. Related names must be different."
            )
    # now we have enough data
    fk = source.meta.fields[foreign_key_name]
    if fk.force_cascade_deletion_relation or (
        fk.on_delete == CASCADE
        and (source.meta.registry is not target.meta.registry or fk.no_constraint)
    ):
        fk.relation_has_post_delete_callback = True
        fk.force_cascade_deletion_relation = True

    related_field = RelatedField(
        foreign_key_name=foreign_key_name,
        name=related_name,
        owner=target,
        related_from=source,
    )

    # Set the related name
    target.meta.fields[related_name] = related_field


def _set_related_name_for_foreign_keys(
    meta: MetaInfo,
    model_class: type[BaseModelType],
    replace_related_field: bool
    | type[BaseModelType]
    | tuple[type[BaseModelType], ...]
    | list[type[BaseModelType]] = False,
) -> None:
    """
    Sets the related name for the foreign keys.
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """
    if not meta.foreign_key_fields:
        return

    for name in meta.foreign_key_fields:
        foreign_key = meta.fields[name]
        related_name = getattr(foreign_key, "related_name", None)
        if related_name is False:
            # skip related_field
            continue

        if not related_name:
            if foreign_key.unique:
                related_name = f"{model_class.__name__.lower()}"
            else:
                related_name = f"{model_class.__name__.lower()}s_set"

        foreign_key.related_name = related_name
        foreign_key.reverse_name = related_name

        related_field_fn = partial(
            _set_related_field,
            source=model_class,
            foreign_key_name=name,
            related_name=related_name,
            replace_related_field=replace_related_field,
        )
        # foreign_key.__dict__.pop("target", None)
        # foreign_key.__dict__.pop("target_registry", None)
        registry: Registry = foreign_key.target_registry
        with contextlib.suppress(Exception):
            registry = cast("Registry", foreign_key.target.registry)
        registry.register_callback(foreign_key.to, related_field_fn, one_time=True)


def _fixup_rel_annotation(target: type[BaseModelType], field: BaseFieldType) -> None:
    if field.is_m2m:
        field.field_type = field.annotation = list[target]  # type: ignore
    elif field.null:
        field.field_type = field.annotation = None | target
    else:
        field.field_type = field.annotation = target


def _fixup_rel_annotations(meta: MetaInfo) -> None:
    for name in chain(meta.foreign_key_fields, meta.many_to_many_fields):
        field = meta.fields[name]
        registry: Registry = field.target_registry
        with contextlib.suppress(Exception):
            registry = cast("Registry", field.target.registry)
        registry.register_callback(
            field.to, partial(_fixup_rel_annotation, field=field), one_time=True
        )


class DatabaseMixin:
    _removed_copy_keys: ClassVar[set[str]] = _removed_copy_keys

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.transaction = self.not_set_transaction

    @classmethod
    def real_add_to_registry(
        cls: type[BaseModelType],
        *,
        registry: Registry,
        registry_type_name: str = "models",
        name: str = "",
        database: bool | Database | Literal["keep"] = "keep",
        replace_related_field: bool
        | type[BaseModelType]
        | tuple[type[BaseModelType], ...]
        | list[type[BaseModelType]] = False,
        on_conflict: Literal["keep", "replace", "error"] = "error",
    ) -> type[BaseModelType]:
        """For customizations."""
        # when called if registry is not set
        cls.meta.registry = registry
        if database is True:
            cls.database = registry.database
        elif database is not False:
            if database == "keep":
                if getattr(cls, "database", None) is None:
                    cls.database = registry.database

            else:
                cls.database = database
        meta = cls.meta
        if name:
            cls.__name__ = name

        # Making sure it does not generate models if abstract or a proxy
        if not meta.abstract and not cls.__is_proxy_model__:
            if on_conflict == "replace":
                registry.delete_model(cls.__name__)
            else:
                with contextlib.suppress(LookupError):
                    original_model = registry.get_model(
                        cls.__name__,
                        include_content_type_attr=False,
                        exclude=("tenant_models",),
                    )
                    if on_conflict == "keep":
                        return original_model
                    else:
                        raise ModelCollisionError(
                            detail=(
                                f'A model with the same name is already registered: "{cls.__name__}".\n'
                                "If this is not a bug, define the behaviour by "
                                'setting "on_conflict" to either "keep" or "replace".'
                            )
                        )
            if registry_type_name:
                registry_dict = getattr(registry, registry_type_name)
                registry_dict[cls.__name__] = cls
                # after registrating the own model
                for value in list(meta.fields.values()):
                    if isinstance(value, BaseManyToManyForeignKeyField):
                        m2m_registry: Registry = value.target_registry
                        with contextlib.suppress(Exception):
                            m2m_registry = cast("Registry", value.target.registry)

                        def create_through_model(x: Any, field: BaseFieldType = value) -> None:
                            # we capture with field = ... the variable
                            field.create_through_model(replace_related_field=replace_related_field)

                        m2m_registry.register_callback(
                            value.to, create_through_model, one_time=True
                        )
                # Sets the foreign key fields
                if meta.foreign_key_fields:
                    _set_related_name_for_foreign_keys(
                        meta, cls, replace_related_field=replace_related_field
                    )
                # fixup annotations required for validation
                if meta.foreign_key_fields or meta.many_to_many_fields:
                    _fixup_rel_annotations(meta)
                registry.execute_model_callbacks(cls)

        # finalize
        cls.model_rebuild(force=True)
        return cls

    @classmethod
    def add_to_registry(
        cls,
        registry: Registry,
        name: str = "",
        database: bool | Database | Literal["keep"] = "keep",
        *,
        replace_related_field: bool
        | type[BaseModelType]
        | tuple[type[BaseModelType], ...]
        | list[type[BaseModelType]] = False,
        on_conflict: Literal["keep", "replace", "error"] = "error",
    ) -> type[BaseModelType]:
        return cls.real_add_to_registry(
            registry=registry,
            name=name,
            database=database,
            replace_related_field=replace_related_field,
            on_conflict=on_conflict,
        )

    def get_active_instance_schema(
        self, check_schema: bool = True, check_tenant: bool = True
    ) -> str | None:
        if self._edgy_namespace["__using_schema__"] is not Undefined:
            return cast(str | None, self._edgy_namespace["__using_schema__"])
        return type(self).get_active_class_schema(
            check_schema=check_schema, check_tenant=check_tenant
        )

    @classmethod
    def get_active_class_schema(cls, check_schema: bool = True, check_tenant: bool = True) -> str:
        if cls.__using_schema__ is not Undefined:
            return cast(str | None, cls.__using_schema__)
        if check_schema:
            schema = get_schema(check_tenant=check_tenant)
            if schema is not None:
                return schema
        db_schema: str | None = cls.get_db_schema()
        # sometime "" is ok, sometimes not, sqlalchemy logic
        return db_schema or None

    @classmethod
    def copy_edgy_model(
        cls: type[Model],
        registry: Registry | None = None,
        name: str = "",
        unlink_same_registry: bool = True,
        on_conflict: Literal["keep", "replace", "error"] = "error",
        **kwargs: Any,
    ) -> type[Self]:
        """Copy the model class and optionally add it to another registry."""
        # removes private pydantic stuff, except the prefixed ones
        attrs = {
            key: val for key, val in cls.__dict__.items() if key not in cls._removed_copy_keys
        }
        # managers and fields are gone, we have to readd them with the correct data
        attrs.update(
            (
                (field_name, field)
                for field_name, field in cls.meta.fields.items()
                if not field.no_copy
            )
        )
        attrs.update(cls.meta.managers)
        _copy = create_edgy_model(
            __name__=name or cls.__name__,
            __module__=cls.__module__,
            __definitions__=attrs,
            __metadata__=cls.meta,
            __bases__=cls.__bases__,
            __type_kwargs__={**kwargs, "skip_registry": True},
        )
        # should also allow masking database with None
        if hasattr(cls, "database"):
            _copy.database = cls.database
        replaceable_models: list[type[BaseModelType]] = [cls]
        for field_name in list(_copy.meta.fields):
            src_field = cls.meta.fields.get(field_name)
            if not isinstance(src_field, BaseForeignKey):
                continue
            # we use the target of source
            replaceable_models.append(src_field.target)

            if src_field.target_registry is cls.meta.registry:
                # clear target_registry, for obvious registries
                del _copy.meta.fields[field_name].target_registry
            if unlink_same_registry and src_field.target_registry is cls.meta.registry:
                # we need to unreference so the target is retrieved from the new registry

                _copy.meta.fields[field_name].target = src_field.target.__name__
            else:
                # otherwise we need to disable backrefs
                _copy.meta.fields[field_name].related_name = False

            if isinstance(src_field, BaseManyToManyForeignKeyField):
                _copy.meta.fields[field_name].through = src_field.through_original
                # clear through registry, we need a copy in the new registry
                del _copy.meta.fields[field_name].through_registry
                if (
                    isinstance(_copy.meta.fields[field_name].through, type)
                    and issubclass(_copy.meta.fields[field_name].through, BaseModelType)
                    and not _copy.meta.fields[field_name].through.meta.abstract
                ):
                    # unreference
                    _copy.meta.fields[field_name].through = through_model = _copy.meta.fields[
                        field_name
                    ].through.copy_edgy_model()
                    # we want to set the registry explicit
                    through_model.meta.registry = False
                    if src_field.from_foreign_key in through_model.meta.fields:
                        # explicit set
                        through_model.meta.fields[src_field.from_foreign_key].target = _copy
                        through_model.meta.fields[src_field.from_foreign_key].related_name = cast(
                            BaseManyToManyForeignKeyField,
                            cast(type[BaseModelType], src_field.through).meta.fields[
                                src_field.from_foreign_key
                            ],
                        ).related_name
        if registry is not None:
            # replace when old class otherwise old references can lead to issues
            _copy.add_to_registry(
                registry,
                replace_related_field=replaceable_models,
                on_conflict=on_conflict,
                database=(
                    "keep"
                    if cls.meta.registry is False or cls.database is not cls.meta.registry.database
                    else True
                ),
            )
        return cast("type[Self]", _copy)

    @property
    def table(self) -> sqlalchemy.Table:
        if self._edgy_namespace.get("_table") is None:
            schema = self.get_active_instance_schema()
            return cast(
                "sqlalchemy.Table",
                type(self).table_schema(schema),
            )
        return cast("sqlalchemy.Table", self._edgy_namespace["_table"])

    @table.setter
    def table(self, value: sqlalchemy.Table | None) -> None:
        self._edgy_namespace.pop("_pkcolumns", None)
        self._edgy_namespace["_table"] = value

    @table.deleter
    def table(self) -> None:
        self._edgy_namespace.pop("_pkcolumns", None)
        self._edgy_namespace.pop("_table", None)

    @property
    def pkcolumns(self) -> Sequence[str]:
        if self._edgy_namespace.get("_pkcolumns") is None:
            if self._edgy_namespace.get("_table") is None:
                self._edgy_namespace["_pkcolumns"] = type(self).pkcolumns
            else:
                self._edgy_namespace["_pkcolumns"] = build_pkcolumns(self)
        return cast(Sequence[str], self._edgy_namespace["_pkcolumns"])

    @property
    def pknames(self) -> Sequence[str]:
        return cast(Sequence[str], type(self).pknames)

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "__using_schema__":
            self._edgy_namespace.pop("_table", None)
        super().__setattr__(key, value)

    def get_columns_for_name(self: Model, name: str) -> Sequence[sqlalchemy.Column]:
        table = self.table
        meta = self.meta
        if name in meta.field_to_columns:
            return meta.field_to_columns[name]
        elif name in table.columns:
            return (table.columns[name],)
        else:
            return cast(Sequence["sqlalchemy.Column"], _empty)

    def identifying_clauses(self, prefix: str = "") -> list[Any]:
        # works only if the class of the model is the main class of the queryset
        # TODO: implement prefix handling and return generic column without table attached
        if prefix:
            raise NotImplementedError()
        clauses: list[Any] = []
        for field_name in self.identifying_db_fields:
            field = self.meta.fields.get(field_name)
            if field is not None:
                for column_name, value in field.clean(
                    field_name, self.__dict__[field_name]
                ).items():
                    clauses.append(getattr(self.table.columns, column_name) == value)
            else:
                clauses.append(
                    getattr(self.table.columns, field_name) == self.__dict__[field_name]
                )
        return clauses

    async def _update(
        self: Model,
        is_partial: bool,
        kwargs: dict[str, Any],
        pre_fn: Callable[..., Awaitable],
        post_fn: Callable[..., Awaitable],
        instance: BaseModelType | QuerySet,
    ) -> int | None:
        """
        Update operation of the database fields.
        """
        real_class = self.get_real_class()
        column_values = self.extract_column_values(
            extracted_values=kwargs,
            is_partial=is_partial,
            is_update=True,
            phase="prepare_update",
            instance=self,
            model_instance=self,
            evaluate_values=True,
        )
        await pre_fn(
            real_class,
            model_instance=self,
            instance=instance,
            values=kwargs,
            column_values=column_values,
        )
        # empty updates shouldn't cause an error. E.g. only model references are updated
        clauses = self.identifying_clauses()
        row_count: int | None = None
        if column_values and clauses:
            check_db_connection(self.database, stacklevel=4)
            async with self.database as database, database.transaction():
                # can update column_values
                column_values.update(
                    await self.execute_pre_save_hooks(column_values, kwargs, is_update=True)
                )
                expression = self.table.update().values(**column_values).where(*clauses)
                row_count = cast(int, await database.execute(expression))

            # Update the model instance.
            new_kwargs = self.transform_input(column_values, phase="post_update", instance=self)
            self.__dict__.update(new_kwargs)

        # updates aren't required to change the db, they can also just affect the meta fields
        await self.execute_post_save_hooks(cast(Sequence[str], kwargs.keys()), is_update=True)

        if column_values or kwargs:
            # Ensure on access refresh the results is active
            self._db_deleted = False if row_count is None else row_count == 0
            self._db_loaded = False
        await post_fn(
            real_class,
            model_instance=self,
            instance=instance,
            values=kwargs,
            column_values=column_values,
        )
        return row_count

    async def update(self: Model, **kwargs: Any) -> Self:
        token = EXPLICIT_SPECIFIED_VALUES.set(set(kwargs.keys()))
        token2 = CURRENT_INSTANCE.set(self)
        try:
            # assume always partial
            await self._update(
                True,
                kwargs,
                pre_fn=partial(
                    self.meta.signals.pre_update.send_async, is_update=True, is_migration=False
                ),
                post_fn=partial(
                    self.meta.signals.post_update.send_async, is_update=True, is_migration=False
                ),
                instance=self,
            )
        finally:
            EXPLICIT_SPECIFIED_VALUES.reset(token)
            CURRENT_INSTANCE.reset(token2)
        return cast("Self", self)

    async def raw_delete(
        self: Model, *, skip_post_delete_hooks: bool, remove_referenced_call: bool | str
    ) -> int:
        """Delete operation from the database"""
        if self._db_deleted:
            return 0
        instance = CURRENT_INSTANCE.get()
        real_class = self.get_real_class()
        # remove_referenced_call = called from a deleter of another field or model
        with_signals = self.__deletion_with_signals__ and (
            instance is not self or remove_referenced_call
        )
        if with_signals:
            await self.meta.signals.pre_delete.send_async(
                real_class, instance=instance, model_instance=self
            )
        ignore_fields: set[str] = set()
        if remove_referenced_call and isinstance(remove_referenced_call, str):
            ignore_fields.add(remove_referenced_call)
        # get values before deleting
        field_values: dict[str, Any] = {}
        if not skip_post_delete_hooks and self.meta.post_delete_fields.difference(ignore_fields):
            token = MODEL_GETATTR_BEHAVIOR.set("coro")
            try:
                for field_name in self.meta.post_delete_fields.difference(ignore_fields):
                    try:
                        field_value = getattr(self, field_name)
                    except AttributeError:
                        # already deleted
                        continue
                    if inspect.isawaitable(field_value):
                        try:
                            field_value = await field_value
                        except AttributeError:
                            # already deleted
                            continue
                    field_values[field_name] = field_value
            finally:
                MODEL_GETATTR_BEHAVIOR.reset(token)
        clauses = self.identifying_clauses()
        row_count = 0
        if clauses:
            expression = self.table.delete().where(*clauses)
            check_db_connection(self.database)
            async with self.database as database:
                row_count = cast(int, await database.execute(expression))
        # we cannot load anymore afterwards
        self._db_deleted = True
        # now cleanup with the saved values
        if field_values:
            token_instance = CURRENT_MODEL_INSTANCE.set(self)
            field_dict: FIELD_CONTEXT_TYPE = cast("FIELD_CONTEXT_TYPE", {})
            token_field_ctx = CURRENT_FIELD_CONTEXT.set(field_dict)
            try:
                for field_name, value in field_values.items():
                    field = self.meta.fields[field_name]
                    field_dict.clear()
                    field_dict["field"] = field
                    await field.post_delete_callback(value)
            finally:
                CURRENT_FIELD_CONTEXT.reset(token_field_ctx)
                CURRENT_MODEL_INSTANCE.reset(token_instance)
        if with_signals:
            await self.meta.signals.post_delete.send_async(
                real_class,
                instance=CURRENT_INSTANCE.get(),
                model_instance=self,
                row_count=row_count,
            )
        return row_count

    async def delete(self: Model, skip_post_delete_hooks: bool = False) -> None:
        """Delete operation from the database"""
        real_class = self.get_real_class()
        await self.meta.signals.pre_delete.send_async(
            real_class, instance=self, model_instance=self
        )
        token = CURRENT_INSTANCE.set(self)
        try:
            row_count = await self.raw_delete(
                skip_post_delete_hooks=skip_post_delete_hooks,
                remove_referenced_call=False,
            )
        finally:
            CURRENT_INSTANCE.reset(token)
        await self.meta.signals.post_delete.send_async(
            real_class, instance=self, model_instance=self, row_count=row_count
        )

    async def load(self, only_needed: bool = False) -> None:
        if only_needed and self._db_loaded_or_deleted:
            return
        row = None
        clauses = self.identifying_clauses()
        if clauses:
            # Build the select expression.
            expression = self.table.select().where(*clauses)

            # Perform the fetch.
            check_db_connection(self.database)
            async with self.database as database:
                row = await database.fetch_one(expression)
        # check if is in system
        if row is None:
            self._db_deleted = True
            self._db_loaded = True
            raise ObjectNotFound("row does not exist anymore")
        # Update the instance.
        self.__dict__.update(self.transform_input(dict(row._mapping), phase="load", instance=self))
        self._db_deleted = False
        self._db_loaded = True

    async def check_exist_in_db(self, only_needed: bool = False) -> bool:
        if only_needed:
            if self._db_deleted:
                return False
            if self._db_loaded:
                return True
        clauses = self.identifying_clauses()
        if not clauses:
            return False

        # Build the select expression.
        expression = self.table.select().where(*clauses).exists().select()

        # Perform the fetch.
        check_db_connection(self.database)
        async with self.database as database:
            result = cast(bool, await database.fetch_val(expression))
            self._db_deleted = not result
            return result

    async def _insert(
        self: Model,
        evaluate_values: bool,
        kwargs: dict[str, Any],
        pre_fn: Callable[..., Awaitable],
        post_fn: Callable[..., Awaitable],
        instance: BaseModelType | QuerySet,
    ) -> None:
        """
        Performs the save instruction.
        """
        real_class = self.get_real_class()
        column_values: dict[str, Any] = self.extract_column_values(
            extracted_values=kwargs,
            is_partial=False,
            is_update=False,
            phase="prepare_insert",
            instance=instance,
            model_instance=self,
            evaluate_values=evaluate_values,
        )
        await pre_fn(
            real_class,
            model_instance=self,
            instance=instance,
            column_values=column_values,
            values=kwargs,
        )
        check_db_connection(self.database, stacklevel=4)
        async with self.database as database, database.transaction():
            # can update column_values
            column_values.update(
                await self.execute_pre_save_hooks(column_values, kwargs, is_update=False)
            )
            expression = self.table.insert().values(**column_values)
            autoincrement_value = await database.execute(expression)
        # sqlalchemy supports only one autoincrement column
        if autoincrement_value:
            column = self.table.autoincrement_column
            if column is not None and hasattr(autoincrement_value, "_mapping"):
                autoincrement_value = autoincrement_value._mapping[column.key]
            # can be explicit set, which causes an invalid value returned
            if column is not None and column.key not in column_values:
                column_values[column.key] = autoincrement_value

        new_kwargs = self.transform_input(column_values, phase="post_insert", instance=self)
        self.__dict__.update(new_kwargs)

        if self.meta.post_save_fields:
            await self.execute_post_save_hooks(cast(Sequence[str], kwargs.keys()), is_update=False)
        # Ensure on access refresh the results is active
        self._db_loaded = False
        self._db_deleted = False
        await post_fn(
            real_class,
            model_instance=self,
            instance=instance,
            column_values=column_values,
            values=kwargs,
        )

    async def real_save(
        self,
        force_insert: bool,
        values: dict[str, Any] | set[str] | None,
    ) -> Self:
        instance: BaseModelType | QuerySet = CURRENT_INSTANCE.get()
        extracted_fields = self.extract_db_fields()
        if values is None:
            explicit_values: set[str] = set()
        elif isinstance(values, set):
            # special mode for marking values as explicit values
            explicit_values = set(values)
            values = None
        else:
            explicit_values = set(values.keys())

        token = MODEL_GETATTR_BEHAVIOR.set("coro")
        try:
            for pkcolumn in type(self).pkcolumns:
                # should trigger load in case of identifying_db_fields
                value = getattr(self, pkcolumn, None)
                if inspect.isawaitable(value):
                    value = await value
                if value is None and self.table.columns[pkcolumn].autoincrement:
                    extracted_fields.pop(pkcolumn, None)
                    force_insert = True
                field = self.meta.fields.get(pkcolumn)
                # this is an IntegerField with primary_key set
                if field is not None and getattr(field, "increment_on_save", 0) != 0:
                    # we create a new revision.
                    force_insert = True
                    # Note: we definitely want this because it is easy for forget a force_insert
            # check if it exists
            if not force_insert and not await self.check_exist_in_db(only_needed=True):
                force_insert = True
        finally:
            MODEL_GETATTR_BEHAVIOR.reset(token)

        token2 = EXPLICIT_SPECIFIED_VALUES.set(explicit_values)
        try:
            if force_insert:
                if values:
                    extracted_fields.update(values)
                # force save must ensure a complete mapping
                await self._insert(
                    bool(values),
                    extracted_fields,
                    pre_fn=partial(
                        self.meta.signals.pre_save.send_async, is_update=False, is_migration=False
                    ),
                    post_fn=partial(
                        self.meta.signals.post_save.send_async, is_update=False, is_migration=False
                    ),
                    instance=instance,
                )
            else:
                await self._update(
                    # assume partial when values are None
                    values is not None,
                    extracted_fields if values is None else values,
                    pre_fn=partial(
                        self.meta.signals.pre_save.send_async, is_update=True, is_migration=False
                    ),
                    post_fn=partial(
                        self.meta.signals.post_save.send_async, is_update=True, is_migration=False
                    ),
                    instance=instance,
                )
        finally:
            EXPLICIT_SPECIFIED_VALUES.reset(token2)
        return self

    async def save(
        self: Model,
        force_insert: bool = False,
        values: dict[str, Any] | set[str] | None = None,
        force_save: bool | None = None,
    ) -> Model:
        """
        Performs a save of a given model instance.
        When creating a user it will make sure it can update existing or
        create a new one.
        """
        if force_save is not None:
            warnings.warn(
                "'force_save' is deprecated in favor of 'force_insert'",
                DeprecationWarning,
                stacklevel=2,
            )
            force_insert = force_save
        token = CURRENT_INSTANCE.set(self)
        try:
            return await self.real_save(force_insert=force_insert, values=values)
        finally:
            CURRENT_INSTANCE.reset(token)

    @classmethod
    def build(
        cls,
        schema: str | None = None,
        metadata: sqlalchemy.MetaData | None = None,
    ) -> sqlalchemy.Table:
        """
        Builds the SQLAlchemy table representation from the loaded fields.
        """
        tablename: str = cls.meta.tablename
        registry = cls.meta.registry
        assert registry, "registry is not set"
        if metadata is None:
            metadata = registry.metadata_by_url[str(cls.database.url)]
        schemes: list[str] = []
        if schema:
            schemes.append(schema)
        if cls.__using_schema__ is not Undefined:
            schemes.append(cls.__using_schema__)
        db_schema = cls.get_db_schema() or ""
        schemes.append(db_schema)

        unique_together = cls.meta.unique_together
        index_constraints = cls.meta.indexes

        columns: list[sqlalchemy.Column] = []
        global_constraints: list[sqlalchemy.Constraint] = [
            copy.copy(constraint) for constraint in cls.meta.constraints
        ]
        for name, field in cls.meta.fields.items():
            current_columns = field.get_columns(name)
            columns.extend(current_columns)
            if not NO_GLOBAL_FIELD_CONSTRAINTS.get():
                global_constraints.extend(
                    field.get_global_constraints(name, current_columns, schemes)
                )

        # Handle the uniqueness together
        uniques = []
        for unique_index in unique_together:
            unique_constraint = cls._get_unique_constraints(unique_index)
            uniques.append(unique_constraint)

        # Handle the indexes
        indexes = []
        for index_c in index_constraints:
            index = cls._get_indexes(index_c)
            indexes.append(index)
        return sqlalchemy.Table(
            tablename,
            metadata,
            *columns,
            *uniques,
            *indexes,
            *global_constraints,
            extend_existing=True,
            schema=(
                schema
                if schema
                else cls.get_active_class_schema(check_schema=False, check_tenant=False)
            ),
        )

    @classmethod
    def add_global_field_constraints(
        cls,
        schema: str | None = None,
        metadata: sqlalchemy.MetaData | None = None,
    ) -> sqlalchemy.Table:
        """
        Add global constraints to table. Required for tenants.
        """
        tablename: str = cls.meta.tablename
        registry = cls.meta.registry
        assert registry, "registry is not set"
        if metadata is None:
            metadata = registry.metadata_by_url[str(cls.database.url)]
        schemes: list[str] = []
        if schema:
            schemes.append(schema)
        if cls.__using_schema__ is not Undefined:
            schemes.append(cls.__using_schema__)
        db_schema = cls.get_db_schema() or ""
        schemes.append(db_schema)
        table = metadata.tables[tablename if not schema else f"{schema}.{tablename}"]
        for name, field in cls.meta.fields.items():
            current_columns: list[sqlalchemy.Column] = []
            for column_name in cls.meta.field_to_column_names[name]:
                current_columns.append(table.columns[column_name])
            for constraint in field.get_global_constraints(name, current_columns, schemes):
                table.append_constraint(constraint)
        return table

    @classmethod
    def _get_unique_constraints(
        cls, fields: Collection[str] | str | UniqueConstraint
    ) -> sqlalchemy.UniqueConstraint | None:
        """
        Returns the unique constraints for the model.

        The columns must be a a list, tuple of strings or a UniqueConstraint object.

        :return: Model UniqueConstraint.
        """
        if isinstance(fields, str):
            return sqlalchemy.UniqueConstraint(
                *cls.meta.field_to_column_names[fields],
                name=hash_names([fields], inner_prefix=cls.__name__, outer_prefix="uc"),
            )
        elif isinstance(fields, UniqueConstraint):
            return sqlalchemy.UniqueConstraint(
                *chain.from_iterable(
                    # deduplicate and extract columns
                    cls.meta.field_to_column_names[field]
                    for field in set(fields.fields)
                ),
                name=fields.name,
                deferrable=fields.deferrable,
                initially=fields.initially,
            )
        # deduplicate
        fields = set(fields)
        return sqlalchemy.UniqueConstraint(
            *chain.from_iterable(cls.meta.field_to_column_names[field] for field in fields),
            name=hash_names(fields, inner_prefix=cls.__name__, outer_prefix="uc"),
        )

    @classmethod
    def _get_indexes(cls, index: Index) -> sqlalchemy.Index | None:
        """
        Creates the index based on the Index fields
        """
        return sqlalchemy.Index(
            index.name,
            *chain.from_iterable(
                (
                    [field]
                    if isinstance(field, sqlalchemy.TextClause)
                    else cls.meta.field_to_column_names[field]
                )
                for field in index.fields
            ),
        )

    def not_set_transaction(self, *, force_rollback: bool = False, **kwargs: Any) -> Transaction:
        """
        Return database transaction for the assigned database.

        This method is automatically assigned to transaction masking the metaclass transaction for instances.
        """
        return cast(
            "Transaction",
            self.database.transaction(force_rollback=force_rollback, **kwargs),
        )



from .base import BasePage, BasePaginator, Page, Paginator
from .cursor import CursorPage, CursorPaginator

__all__ = ["BasePage", "BasePaginator", "Page", "Paginator", "CursorPage", "CursorPaginator"]



from __future__ import annotations

import sys
from typing import Any, ClassVar, cast

from pydantic import ConfigDict

from edgy.core.db.models.base import EdgyBaseModel
from edgy.core.db.models.managers import BaseManager, Manager, RedirectManager
from edgy.core.db.models.metaclasses import BaseModelMeta, MetaInfo
from edgy.core.db.models.mixins.admin import AdminMixin
from edgy.core.db.models.mixins.db import DatabaseMixin
from edgy.core.db.models.mixins.dump import DumpMixin
from edgy.core.db.models.mixins.generics import DeclarativeMixin
from edgy.core.db.models.mixins.reflection import ReflectedModelMixin
from edgy.core.db.models.mixins.row import ModelRowMixin
from edgy.core.utils.models import create_edgy_model, generify_model_fields

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self


class Model(
    ModelRowMixin,
    DeclarativeMixin,
    DatabaseMixin,
    AdminMixin,
    DumpMixin,
    EdgyBaseModel,
    metaclass=BaseModelMeta,
):
    """
    Representation of an Edgy `Model`.

    This also means it can generate declarative SQLAlchemy models
    from anywhere by calling the `Model.declarative()` function.

    **Example**

    ```python
    import edgyBaseFieldType
    from edgy import Database, Registry

    database = Database("sqlite:///db.sqlite")
    models = Registry(database=database)


    class User(edgy.Model):
        '''
        The User model to be created in the database as a table
        If no name is provided the in Meta class, it will generate
        a "users" table for you.
        '''

        id: int = edgy.IntegerField(primary_key=True, autoincrement=True)
        is_active: bool = edgy.BooleanField(default=False)

        class Meta:
            registry = models
    ```
    """

    # update typings
    proxy_model: ClassVar[type[Self]]
    __parent__: ClassVar[type[Self] | None]
    query: ClassVar[BaseManager] = Manager()
    query_related: ClassVar[BaseManager] = RedirectManager(redirect_name="query")
    # registry = False, stops the retrieval of the registry from base classes
    meta: ClassVar[MetaInfo] = MetaInfo(None, abstract=True, registry=False)

    class Meta:
        abstract = True
        # registry = False, stops the retrieval of the registry from base classes
        registry = False

    @classmethod
    def real_add_to_registry(cls, **kwargs: Any) -> type[Model]:
        in_admin = cls.meta.in_admin
        result = cast(type[Model], super().real_add_to_registry(**kwargs))
        if in_admin is not False and result.meta.registry:
            result.meta.registry.admin_models.add(result.__name__)
        return result

    @classmethod
    def generate_proxy_model(cls) -> type[Model]:
        """
        Generates a proxy model for each model. This proxy model is a simple
        shallow copy of the original model being generated.
        """
        # a proxy model is a copy of the model but without replaced field references and is not added to registry
        # it can't either

        # removes private pydantic stuff, except the prefixed ones
        attrs: dict[str, Any] = {
            key: val for key, val in cls.__dict__.items() if key not in cls._removed_copy_keys
        }
        # managers and fields are gone, we have to readd them with the correct data
        # Note: because it is not a real copy don't honor no_copy flags.
        attrs.update(cls.meta.fields)
        attrs.update(cls.meta.managers)
        attrs["__is_proxy_model__"] = True
        _copy = create_edgy_model(
            __name__=cls.__name__,
            __module__=cls.__module__,
            __definitions__=attrs,
            __metadata__=cls.meta,
            __bases__=cls.__bases__,
            # registry is still added because it is copied from meta
            __type_kwargs__={"skip_registry": True},
        )
        _copy.database = cls.database
        generify_model_fields(_copy)
        return _copy


class StrictModel(Model):
    model_config = ConfigDict(
        extra="forbid", arbitrary_types_allowed=True, validate_assignment=True, strict=True
    )

    class Meta:
        abstract = True
        # registry = False, stops the retrieval of the registry from base classes
        registry = False


class ReflectModel(ReflectedModelMixin, Model):
    """
    Reflect on async engines is not yet supported, therefore, we need to make a sync_engine
    call.
    """

    class Meta:
        abstract = True
        # registry = False, stops the retrieval of the registry from base classes
        registry = False



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from edgy.core.db.models import Model


class ConfigMarshall(TypedDict, total=False):
    """A TypedDict for configuring Marshall behaviour."""

    model: type[Model] | str
    """The model from there the marshall will read from."""

    fields: list[str] | None = None
    """A list of fields to be serialized"""

    exclude: list[str] | None = None
    """A list of fields to be excluded from the serialization."""



from __future__ import annotations

from collections.abc import AsyncGenerator, Hashable, Iterable
from typing import TYPE_CHECKING, Any

from .base import BasePage, BasePaginator

if TYPE_CHECKING:
    from edgy.core.db.models.types import BaseModelType
    from edgy.core.db.querysets import QuerySet


class CursorPage(BasePage):
    next_cursor: Hashable | None
    current_cursor: Hashable | None


class CursorPaginator(BasePaginator[CursorPage]):
    def __init__(
        self,
        queryset: QuerySet,
        page_size: int,
        next_item_attr: str = "",
        previous_item_attr: str = "",
    ) -> None:
        super().__init__(
            queryset=queryset,
            page_size=page_size,
            next_item_attr=next_item_attr,
            previous_item_attr=previous_item_attr,
        )
        self._reverse_page_cache: dict[Hashable, CursorPage] = {}
        self.search_vector = self.calculate_search_vector()

    def calculate_search_vector(self) -> tuple[str, ...]:
        vector = [
            f"{criteria[1:]}__lte" if criteria.startswith("-") else f"{criteria}__gte"
            for criteria in self.order_by[:-1]
        ]
        criteria = self.order_by[-1]
        vector.append(f"{criteria[1:]}__lt" if criteria.startswith("-") else f"{criteria}__gt")

        return tuple(vector)

    def cursor_to_vector(self, cursor: Hashable) -> tuple[Hashable, ...]:
        if isinstance(cursor, tuple):
            return cursor
        assert len(self.order_by) == 1
        return (cursor,)

    def vector_to_cursor(self, vector: tuple[Hashable, ...]) -> Hashable:
        if len(self.order_by) > 1:
            return vector
        return vector[0]

    def obj_to_vector(self, obj: Any) -> tuple:
        return tuple(getattr(obj, attr.lstrip("-")) for attr in self.order_by)

    def obj_to_cursor(self, obj: Any) -> Hashable:
        return self.vector_to_cursor(self.obj_to_vector(obj))

    def clear_caches(self) -> None:
        super().clear_caches()
        self._reverse_page_cache.clear()

    def convert_to_page(
        self, inp: Iterable, /, cursor: Hashable | None, is_first: bool, reverse: bool = False
    ) -> CursorPage:
        page_obj: BasePage = super().convert_to_page(
            inp,
            is_first=is_first,
            reverse=reverse,
        )
        next_cursor = self.obj_to_cursor(page_obj.content[-1]) if page_obj.content else None
        return CursorPage(
            content=page_obj.content,
            is_first=page_obj.is_first,
            is_last=page_obj.is_last,
            next_cursor=next_cursor,
            current_cursor=cursor,
        )

    async def get_extra_before(self, cursor: Hashable) -> list[BaseModelType]:
        vector = self.cursor_to_vector(cursor)
        rpaginator = self.get_reverse_paginator()
        return await rpaginator.queryset.filter(
            **dict(zip(rpaginator.search_vector, vector, strict=False))
        ).limit(1)

    async def exists_extra_before(self, cursor: Hashable) -> bool:
        vector = self.cursor_to_vector(cursor)
        rpaginator = self.get_reverse_paginator()
        return await rpaginator.queryset.filter(
            **dict(zip(rpaginator.search_vector, vector, strict=False))
        ).exists()

    async def _get_page_after(
        self,
        vector: tuple[Hashable, ...] | None,
        injected_extra: list[BaseModelType] | None = None,
        reverse: bool = False,
    ) -> tuple[CursorPage, list[BaseModelType]]:
        query = self.queryset.limit(self.page_size + 1) if self.page_size else self.queryset
        if vector is not None:
            query = query.filter(**dict(zip(self.search_vector, vector, strict=False)))
        is_first = vector is None
        if not is_first and self.previous_item_attr:
            resultarr = (
                await self.get_extra_before(vector) if injected_extra is None else injected_extra
            )
            # if on first position
            if not resultarr:
                is_first = True
            resultarr.extend(await query)
        elif injected_extra is not None:
            if injected_extra:
                is_first = False
            injected_extra.extend(await query)
            resultarr = injected_extra
        else:
            if not is_first and not await self.exists_extra_before(vector):
                is_first = True
            resultarr = await query

        page_obj = self.convert_to_page(
            resultarr,
            cursor=None if vector is None else self.vector_to_cursor(vector),
            is_first=is_first,
            reverse=reverse,
        )
        return page_obj, resultarr

    async def get_page_after(self, cursor: Hashable | None = None) -> CursorPage:
        vector: tuple[Hashable, ...] | None = None
        if cursor is not None:
            vector = self.cursor_to_vector(cursor)
        if vector in self._page_cache:
            page_obj = self._page_cache[vector]
            return page_obj
        page_obj = (await self._get_page_after(vector=vector))[0]
        self._page_cache[vector] = page_obj
        return page_obj

    async def get_page_before(self, cursor: Hashable | None = None) -> CursorPage:
        vector: tuple[Hashable, ...] | None = None
        if cursor is not None:
            vector = self.cursor_to_vector(cursor)
        if vector in self._reverse_page_cache:
            page_obj = self._reverse_page_cache[vector]
            return page_obj
        reverse_paginator = self.get_reverse_paginator()
        new_vector: tuple[Hashable, ...] | None = None
        # to match the cursors we need to go back one more item
        injected_reverse = (
            await reverse_paginator.get_extra_before(vector) if vector is not None else []
        )
        if injected_reverse:
            new_vector = self.obj_to_vector(injected_reverse[0])
        # instead of recalculating get_extra_before inject the item
        # we need the extra item in the array (is_last) if it exists, so we match the cursor of
        # get_page_after
        reverse_page, raw_array = await reverse_paginator._get_page_after(
            new_vector, injected_extra=injected_reverse, reverse=True
        )
        page_obj = CursorPage(
            content=reverse_page.content,
            is_first=reverse_page.is_first,
            is_last=reverse_page.is_last,
            next_cursor=self.obj_to_cursor(raw_array[-1]) if raw_array else None,
            current_cursor=None if vector is None else self.vector_to_cursor(vector),
        )
        self._reverse_page_cache[vector] = reverse_page
        return page_obj

    async def get_page(self, cursor: Hashable | None = None, backward: bool = False) -> CursorPage:
        # this reverse only reverses the direction in which the cursor is evaluated
        if backward:
            return await self.get_page_before(cursor)
        else:
            return await self.get_page_after(cursor)

    async def paginate(
        self, start_cursor: Hashable | None = None, stop_cursor: Hashable | None = None
    ) -> AsyncGenerator[CursorPage, None]:
        query = self.queryset
        prefill_container = []
        start_vector: tuple[Hashable, ...] | None = None
        if start_cursor is not None:
            start_vector = self.cursor_to_vector(start_cursor)
            query = query.filter(**dict(zip(self.search_vector, start_vector, strict=False)))
            if self.previous_item_attr:
                prefill_container = await self.get_extra_before(start_vector)
        if stop_cursor is not None:
            stop_vector = self.cursor_to_vector(stop_cursor)
            query = query.filter(
                **dict(zip(self.get_reverse_paginator().search_vector, stop_vector, strict=False))
            )
        current_cursor: Hashable | None = (
            None if start_vector is None else self.vector_to_cursor(start_vector)
        )
        async for page in self.paginate_queryset(
            query,
            is_first=bool(start_cursor is None and not prefill_container),
            prefill=prefill_container,
        ):
            next_cursor = self.obj_to_cursor(page.content[-1]) if page.content else None
            yield CursorPage(
                **page.__dict__, next_cursor=next_cursor, current_cursor=current_cursor
            )
            current_cursor = next_cursor



from __future__ import annotations

import inspect
import sys
from asyncio import gather
from collections.abc import Awaitable
from functools import cached_property
from typing import TYPE_CHECKING, Any, ClassVar, cast

from pydantic import BaseModel, ConfigDict, Field
from pydantic.fields import FieldInfo
from pydantic.json_schema import SkipJsonSchema

from edgy.core.db.models.mixins.dump import DumpMixin
from edgy.core.marshalls.config import ConfigMarshall
from edgy.core.marshalls.fields import BaseMarshallField
from edgy.core.marshalls.metaclasses import MarshallMeta
from edgy.core.utils.sync import run_sync

if TYPE_CHECKING:
    from edgy.core.db.models.metaclasses import MetaInfo
    from edgy.core.db.models.model import Model


if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self

excludes_marshall: set = {"instance", "_instance", "context"}


class BaseMarshall(DumpMixin, BaseModel):
    """
    Base for all the marshalls of Edgy.
    """

    marshall_config: ClassVar[ConfigMarshall]
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

    __show_pk__: ClassVar[bool] = False
    __lazy__: ClassVar[bool] = False
    __incomplete_fields__: ClassVar[tuple[str, ...]] = ()
    __custom_fields__: ClassVar[dict[str, BaseMarshallField]] = {}
    _setup_used: bool

    def __init__(self, instance: None | Model = None, **kwargs: Any) -> None:
        lazy = kwargs.pop("__lazy__", type(self).__lazy__)
        data: dict = {}
        if instance is not None:
            data.update(
                instance.model_dump(
                    exclude_defaults=True,
                    exclude_unset=True,
                    exclude=excludes_marshall.union(self.__custom_fields__),
                )
            )
        data.update(kwargs)
        super().__init__(**data)
        self._instance: Model | None = None
        if instance is not None:
            self.instance = instance
        elif not lazy:
            self._instance = self._setup()
            self._resolve_serializer(self._instance)
            self._setup_used = True

    def _setup(self) -> Model:
        """
        Assemble the Marshall object with all the given details.

        Returns:
            Model or None: The assembled model instance, or None if the assembly fails.
        """
        klass = type(self)
        if klass.__incomplete_fields__:
            raise RuntimeError(
                f"'{klass.__name__}' is an incomplete Marshall. "
                f"For creating new instances, it lacks following fields: [{', '.join(klass.__incomplete_fields__)}]."
            )
        model = cast("Model", self.marshall_config["model"])
        column = model.table.autoincrement_column
        exclude: set[str] = {*excludes_marshall}
        if column is not None:
            exclude.add(column.key)
        data = self.model_dump(
            include=set(model.meta.fields.keys()).difference(exclude),
        )
        for k in list(data.keys()):
            # callable defaults are not handled in marshall but can leak from models fields
            if callable(data[k]):
                data.pop(k)

        data["__show_pk__"] = self.__show_pk__
        data["__drop_extra_kwargs__"] = True
        return self.marshall_config["model"](**data)  # type: ignore

    @property
    def meta(self) -> MetaInfo:
        return cast("Model", self.marshall_config["model"]).meta

    @property
    def has_instance(self) -> bool:
        return self._instance is not None

    @property
    def instance(self) -> Model:
        if self._instance is None:
            _instance = self._setup()
            self._resolve_serializer(_instance)
            self._instance = _instance
            self._setup_used = True
        return self._instance

    @instance.setter
    def instance(self, value: Model) -> None:
        self._instance = value
        self._setup_used = False
        self._resolve_serializer(instance=value)

    async def _resolve_async(self, name: str, awaitable: Awaitable) -> None:
        setattr(self, name, await awaitable)

    def _resolve_serializer(self, instance: Model) -> Self:
        """
        Resolve serializer fields and populate them with data from the provided instance.

        Args:
            instance (Model): The instance to extract data from.

        Returns:
            BaseMarshall: The resolved serializer instance.
        """
        # Dump model data excluding fields extracted above
        async_resolvers = []
        # Iterate over fields to populate them with data
        for name, field in self.__custom_fields__.items():
            if not field.__is_method__:
                # For primary key exceptions
                if name in instance.pknames:
                    attribute = getattr(instance, name)
                else:
                    attribute = getattr(instance, field.source or name)

                # If attribute is callable, execute it and set the value
                if callable(attribute):
                    value = attribute()
                    if inspect.isawaitable(value):
                        async_resolvers.append(self._resolve_async(name, value))
                        continue
                else:
                    value = attribute
                setattr(self, name, value)
            elif field.__is_method__:
                value = self._get_method_value(name, instance)
                # allow async getters
                if inspect.isawaitable(value):
                    async_resolvers.append(self._resolve_async(name, value))
                    continue
                setattr(self, name, value)
        if async_resolvers:
            run_sync(gather(*async_resolvers))
        return self

    def _get_method_value(self, name: str, instance: Model) -> Any:
        """
        Retrieve the value for a method-based field.

        Args:
            name (str): The name of the method-based field.
            instance (Model): The instance to retrieve the value from.

        Returns:
            Optional: The value retrieved from the method-based field.
        """
        func_name: str = f"get_{name}"
        func = getattr(self, func_name)
        return func(instance)

    def _handle_primary_key(self, instance: Model) -> None:
        """
        Handles the field of a the primary key if present in the
        fields.
        """
        data = self.model_dump(include=set(instance.pknames))

        # fix incomplete data
        if data:
            for pk_attribute in instance.pknames:
                # bypass __setattr__ method
                object.__setattr__(self, pk_attribute, getattr(instance, pk_attribute))

    @cached_property
    def valid_fields(self) -> dict[str, FieldInfo]:
        """
        Returns all valid fields (not excluded). Overwritable and can be manually set.
        """
        return {
            k: v for k, v in type(self).model_fields.items() if not getattr(v, "exclude", True)
        }

    @cached_property
    def fields(self) -> dict[str, BaseMarshallField]:
        """
        Returns all the Marshall fields of the Marshall. Overwritable and can be manually set.
        """
        fields = {}
        # copy fields from model_fields
        for k, v in type(self).model_fields.items():
            if isinstance(v, BaseMarshallField):
                fields[k] = v
        return fields

    async def save(self) -> BaseMarshall:
        """
        This save is not actually performing any specific action,
        in fact, its using the Edgy saving method and nothing else.

        We add the save here to make sure we have compatibility with
        the interface of Edgy.

        This is optional! This should only be called if you indeed
        want to persist the object in the database and have access to a
        primary key.

        !!! Tip
            All the field and model validations **are still performed**
            in the model level, not in a marshall level.
        """
        model = cast("Model", self.marshall_config["model"])
        if self._setup_used:
            # use defaults of Marshall, save completely
            instance = await self.instance.save()
        else:
            column = model.table.autoincrement_column
            exclude: set[str] = set()
            if column is not None:
                exclude.add(column.key)
            data = self.model_dump(
                include=set(model.meta.fields.keys()).difference(exclude), exclude_unset=True
            )
            # update without using defaults
            instance = await self.instance.save(values=data)
        self._handle_primary_key(instance=instance)
        return self


class Marshall(BaseMarshall, metaclass=MarshallMeta):
    """
    Model marshall where the `__model__` is required.
    """

    context: SkipJsonSchema[dict] = Field(exclude=True, default_factory=dict)

    async def save(self) -> Self:
        return cast("Self", await super().save())

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self}>"

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}({cast('type[Model]', self.marshall_config['model']).__name__})"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast

from lilya.context import session
from pydantic.json_schema import GenerateJsonSchema, NoDefault
from pydantic_core import core_schema

import edgy

if TYPE_CHECKING:
    from edgy.core.db.models.model import Model


class CallableDefaultJsonSchema(GenerateJsonSchema):
    def get_default_value(self, schema: core_schema.WithDefaultSchema) -> Any:
        value = super().get_default_value(schema)
        if callable(value):
            value = value()
        return value


class NoCallableDefaultJsonSchema(GenerateJsonSchema):
    def get_default_value(self, schema: core_schema.WithDefaultSchema) -> Any:
        value = super().get_default_value(schema)
        if callable(value):
            value = NoDefault
        return value


def get_registered_models() -> dict[str, type[Model]]:
    registry = edgy.monkay.instance.registry
    return {name: registry.get_model(name) for name in registry.admin_models}


def get_model(model_name: str, *, no_check_admin_models: bool = False) -> type[Model]:
    registry = edgy.monkay.instance.registry
    if not no_check_admin_models and model_name not in registry.admin_models:
        raise LookupError()
    return cast("type[Model]", registry.get_model(model_name, exclude={"pattern_models"}))


def get_model_json_schema(
    model: str | type[Model],
    /,
    mode: Literal["validation", "serialization"] = "validation",
    phase: str = "view",
    include_callable_defaults: bool = False,
    no_check_admin_models: bool = False,
    **kwargs: Any,
) -> dict:
    if isinstance(model, str):
        model = get_model(model, no_check_admin_models=no_check_admin_models)
    marshall_class = model.get_admin_marshall_class(phase=phase, for_schema=True)
    return marshall_class.model_json_schema(
        schema_generator=CallableDefaultJsonSchema
        if include_callable_defaults
        else NoCallableDefaultJsonSchema,
        mode=mode,
        **kwargs,
    )


def add_to_recent_models(model: type[Model]) -> None:
    if not model.meta.registry:
        return
    if hasattr(session, "recent_models"):
        recent_models = [name for name in session.recent_models[:10] if name != model.__name__]
    else:
        recent_models = []
    recent_models.insert(0, model.__name__)
    session.recent_models = recent_models


def get_recent_models() -> list[str]:
    if not getattr(session, "recent_models", None):
        return []
    return cast(list[str], session.recent_models)
